from flask import Blueprint

from argus.backend.plugins.sct.testrun import SCTTestRun
from argus.backend.plugins.sct.controller import bp as sct_bp
from argus.backend.plugins.core import PluginInfoBase, PluginModelBase
from argus.backend.plugins.sct.udt import (
    CloudInstanceDetails,
    CloudNodesInfo,
    CloudResource,
    CloudSetupDetails,
    EventsBySeverity,
    NemesisRunInfo,
    NodeDescription,
    PackageVersion,
    PerformanceHDRHistogram,
)


class PluginInfo(PluginInfoBase):
    # pylint: disable=too-few-public-methods
    name: str = "scylla-cluster-tests"
    model: PluginModelBase = SCTTestRun
    controller: Blueprint = sct_bp
    all_models = [
        SCTTestRun,
    ]
    all_types = [
        NemesisRunInfo,
        NodeDescription,
        EventsBySeverity,
        CloudResource,
        CloudSetupDetails,
        CloudNodesInfo,
        CloudInstanceDetails,
        PackageVersion,
        PerformanceHDRHistogram,
    ]



import logging
from flask import (
    Blueprint,
    render_template,
)
from argus.backend.controller.admin_api import bp as admin_api_bp
from argus.backend.service.user import login_required, check_roles
from argus.backend.models.web import UserRoles

bp = Blueprint('admin', __name__, url_prefix='/admin')
bp.register_blueprint(admin_api_bp)
LOGGER = logging.getLogger(__name__)


@bp.route("/")
@bp.route("/<string:path>")
@check_roles(UserRoles.Admin)
@login_required
def index(path: str = "index"):
    return render_template("admin/index.html.j2", current_route=path)



from dataclasses import dataclass
from functools import reduce
import logging
import math
from time import time
from flask import g
from argus.backend.models.web import ArgusEventTypes
from argus.backend.plugins.sct.testrun import SCTTestRun, SubtestType
from argus.backend.plugins.sct.types import GeminiResultsRequest, PerformanceResultsRequest
from argus.backend.plugins.sct.udt import (
    CloudInstanceDetails,
    CloudResource,
    EventsBySeverity,
    NemesisRunInfo,
    NodeDescription,
    PackageVersion,
    PerformanceHDRHistogram,
)
from argus.backend.service.event_service import EventService
from argus.backend.util.common import get_build_number
from argus.backend.util.enums import NemesisStatus, ResourceState, TestStatus

LOGGER = logging.getLogger(__name__)


class SCTServiceException(Exception):
    pass


@dataclass(init=True, repr=True)
class NemesisSubmissionRequest:
    name: str
    class_name: str
    start_time: int
    node_name: str
    node_ip: str
    node_shards: int


@dataclass(init=True, repr=True)
class NemesisFinalizationRequest:
    name: str
    start_time: int
    status: str
    message: str


@dataclass(init=True, repr=True)
class EventSubmissionRequest:
    severity: str
    total_events: int
    messages: list[str]


class SCTService:

    @staticmethod
    def submit_packages(run_id: str, packages: list[dict]) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            for package_dict in packages:
                package = PackageVersion(**package_dict)
                run.packages.append(package)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "added"


    @staticmethod
    def set_sct_runner(run_id: str, public_ip: str, private_ip: str, region: str, backend: str):
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.sct_runner_host = CloudInstanceDetails(
                public_ip=public_ip,
                private_ip=private_ip,
                provider=backend,
                region=region,
            )
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "updated"

    @staticmethod
    def submit_screenshots(run_id: str, screenshot_links: list[str]) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            for link in screenshot_links:
                run.add_screenshot(link)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "submitted"

    @staticmethod
    def submit_gemini_results(run_id: str, gemini_data: GeminiResultsRequest) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.subtest_name = SubtestType.GEMINI.value
            run.oracle_nodes_count = gemini_data.get("oracle_nodes_count")
            run.oracle_node_ami_id = gemini_data.get("oracle_node_ami_id")
            run.oracle_node_instance_type = gemini_data.get("oracle_node_instance_type")
            run.oracle_node_scylla_version = gemini_data.get("oracle_node_scylla_version")
            run.gemini_command = gemini_data.get("gemini_command")
            run.gemini_version = gemini_data.get("gemini_version")
            run.gemini_status = gemini_data.get("gemini_status")
            run.gemini_seed = str(gemini_data.get("gemini_seed"))
            run.gemini_write_ops = gemini_data.get("gemini_write_ops")
            run.gemini_write_errors = gemini_data.get("gemini_write_errors")
            run.gemini_read_ops = gemini_data.get("gemini_read_ops")
            run.gemini_read_errors = gemini_data.get("gemini_read_errors")
            run.save()

            if run.gemini_status != "PASSED":
                run.status = TestStatus.FAILED
                EventService.create_run_event(kind=ArgusEventTypes.TestRunStatusChanged, body={
                        "message": "[{username}] Setting run status to {status} due to Gemini reporting following status: {gemini_status}",
                        "username": g.user.username,
                        "status": TestStatus.FAILED.value,
                        "gemini_status": run.gemini_status,
                }, user_id=g.user.id, run_id=run_id, release_id=run.release_id, test_id=run.test_id)
                run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "submitted"

    @staticmethod
    def submit_performance_results(run_id: str, performance_results: PerformanceResultsRequest):
        # pylint: disable=too-many-statements
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.subtest_name = SubtestType.PERFORMANCE.value
            run.perf_op_rate_average = performance_results.get("perf_op_rate_average")
            run.perf_op_rate_total = performance_results.get("perf_op_rate_total")
            run.perf_avg_latency_99th = performance_results.get("perf_avg_latency_99th")
            run.perf_avg_latency_mean = performance_results.get("perf_avg_latency_mean")
            run.perf_total_errors = performance_results.get("perf_total_errors")
            run.stress_cmd = performance_results.get("stress_cmd")
            run.test_name = performance_results.get("test_name")
            run.save()

            is_latency_test = "latency" in run.test_name
            threshold_negative = -10

            def cmp(lhs, rhs):
                delta = rhs - lhs
                change = int(math.fabs(delta) * 100 / rhs)
                return change if delta >= 0 else change * -1

            previous_runs = SCTTestRun.get_perf_results_for_test_name(run.build_id, run.start_time, run.test_name)
            metrics_to_check = ["perf_avg_latency_99th", "perf_avg_latency_mean"] if is_latency_test else ["perf_op_rate_total"]

            older_runs_by_version = {}
            for prev_run in previous_runs:
                if not older_runs_by_version.get(prev_run["scylla_version"]):
                    older_runs_by_version[prev_run["scylla_version"]] = []
                older_runs_by_version[prev_run["scylla_version"]].append(prev_run)

            regression_found = False
            regression_info = {
                "version": None,
                "delta": None,
                "id": None,
                "metric": None,
                "job_url": None,
            }

            if performance_results["histograms"]:
                for histogram in performance_results["histograms"]:
                    run.histograms = { k: PerformanceHDRHistogram(**v) for k, v in histogram.items() }

            for version, runs in older_runs_by_version.items():
                for metric in metrics_to_check:
                    # pylint: disable=cell-var-from-loop
                    best_run = sorted(runs, reverse=(not is_latency_test), key=lambda v: v[metric])[0]
                    last_run = runs[0]

                    metric_to_best = cmp(run[metric], best_run[metric])
                    metric_to_last = cmp(run[metric], last_run[metric])
                    if metric_to_last < threshold_negative:
                        regression_found = True
                        regression_info["metric"] = metric
                        regression_info["version"] = version
                        regression_info["job_url"] = last_run["build_job_url"]
                        regression_info["id"] = str(last_run["id"])
                        regression_info["delta"] = metric_to_last
                        break

                    if metric_to_best < threshold_negative:
                        regression_found = True
                        regression_info["metric"] = metric
                        regression_info["version"] = version
                        regression_info["job_url"] = best_run["build_job_url"]
                        regression_info["id"] = str(best_run["id"])
                        regression_info["delta"] = metric_to_best
                        break

                if regression_found:
                    break

            if regression_found:
                run.status = TestStatus.FAILED.value
                run.save()
                EventService.create_run_event(kind=ArgusEventTypes.TestRunStatusChanged, body={
                        "message": "[{username}] Setting run status to {status} due to performance metric '{metric}' falling "
                                   "below allowed threshold ({threshold_negative}): {delta}% compared to "
                                   "<a href='/test/{test_id}/runs?additionalRuns[]={base_run_id}&additionalRuns[]={previous_run_id}'>This {version} (#{build_number}) run</a>",
                        "username": g.user.username,
                        "status": TestStatus.FAILED.value,
                        "metric": regression_info["metric"],
                        "threshold_negative": threshold_negative,
                        "delta": regression_info["delta"],
                        "test_id": str(run.test_id),
                        "base_run_id": str(run.id),
                        "previous_run_id": regression_info["id"],
                        "version": regression_info["version"],
                        "build_number": get_build_number(regression_info["job_url"])
                }, user_id=g.user.id, run_id=run_id, release_id=run.release_id, test_id=run.test_id)
            else:
                # NOTE: This will override status set by SCT Events.
                run.status = TestStatus.PASSED.value
                run.save()

        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "submitted"

    @staticmethod
    def get_performance_history_for_test(run_id: str):
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            rows = run.get_perf_results_for_test_name(build_id=run.build_id, start_time=run.start_time, test_name=run.test_name)
            return rows
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception


    @staticmethod
    def create_resource(run_id: str, resource_details: dict) -> str:
        instance_details = CloudInstanceDetails(**resource_details.pop("instance_details"))
        resource = CloudResource(**resource_details, instance_info=instance_details)
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.get_resources().append(resource)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "created"

    @staticmethod
    def update_resource_shards(run_id: str, resource_name: str, new_shards: int) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            resource = next(res for res in run.get_resources() if res.name == resource_name)
            resource.get_instance_info().shards_amount = new_shards
            run.save()
        except StopIteration as exception:
            LOGGER.error("Resource %s not found in run %s", resource_name, run_id)
            raise SCTServiceException("Resource not found", resource_name) from exception
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "updated"

    @staticmethod
    def terminate_resource(run_id: str, resource_name: str, reason: str) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            resource = next(res for res in run.get_resources() if res.name == resource_name)
            resource.get_instance_info().termination_reason = reason
            resource.get_instance_info().termination_time = int(time())
            resource.state = ResourceState.TERMINATED.value
            run.save()
        except StopIteration as exception:
            LOGGER.error("Resource %s not found in run %s", resource_name, run_id)
            raise SCTServiceException("Resource not found", resource_name) from exception
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "terminated"

    @staticmethod
    def submit_nemesis(run_id: str, nemesis_details: dict) -> str:
        nem_req = NemesisSubmissionRequest(**nemesis_details)
        node_desc = NodeDescription(name=nem_req.node_name, ip=nem_req.node_ip, shards=nem_req.node_shards)
        nemesis_info = NemesisRunInfo(
            class_name=nem_req.class_name,
            name=nem_req.name,
            start_time=int(nem_req.start_time),
            end_time=0,
            duration=0,
            stack_trace="",
            status=NemesisStatus.RUNNING.value,
            target_node=node_desc,
        )
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.add_nemesis(nemesis_info)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "created"

    @staticmethod
    def finalize_nemesis(run_id: str, nemesis_details: dict) -> str:
        nem_req = NemesisFinalizationRequest(**nemesis_details)
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            nemesis = next(nem for nem in run.get_nemeses() if nem.name ==
                           nem_req.name and nem.start_time == nem_req.start_time)
            nemesis.status = NemesisStatus(nem_req.status).value
            nemesis.stack_trace = nem_req.message
            nemesis.end_time = int(time())
            run.save()
        except StopIteration as exception:
            LOGGER.error("Nemesis %s (%s) not found for run %s", nem_req.name, nem_req.start_time, run_id)
            raise SCTServiceException("Nemesis not found", (nem_req.name, nem_req.start_time)) from exception
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "updated"

    @staticmethod
    def submit_events(run_id: str, events: list[dict]) -> str:
        wrapped_events = [EventSubmissionRequest(**ev) for ev in events]
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            for event in wrapped_events:
                wrapper = EventsBySeverity(severity=event.severity,
                                           event_amount=event.total_events, last_events=event.messages)
                run.get_events().append(wrapper)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "added"

    @staticmethod
    def get_scylla_version_kernels_report(release_name: str):
        all_release_runs = SCTTestRun.get_version_data_for_release(release_name=release_name)
        kernels_by_version = {}
        kernel_metadata = {}
        for run in all_release_runs:
            packages = run["packages"]
            if not packages:
                continue
            scylla_pkgs = {p["name"]: p for p in packages if "scylla-server" in p["name"]}
            scylla_pkg = scylla_pkgs["scylla-server-upgraded"] if scylla_pkgs.get(
                "scylla-server-upgraded") else scylla_pkgs.get("scylla-server")
            version = f"{scylla_pkg['version']}-{scylla_pkg['date']}.{scylla_pkg['revision_id']}" if scylla_pkgs else "unknown"
            kernel_packages = [p for p in packages if "kernel" in p["name"]]
            kernel_package = kernel_packages[0] if len(kernel_packages) > 0 else None
            if not kernel_package:
                continue
            version_list = set(kernels_by_version.get(version, []))
            version_list.add(kernel_package["version"])
            kernels_by_version[version] = list(version_list)
            metadata = kernel_metadata.get(
                kernel_package.version,
                {
                    "passed": 0,
                    "failed": 0,
                    "aborted": 0,
                }
            )
            if run["status"] in ["passed", "failed", "aborted", "test_error"]:
                metadata[run["status"]] += 1
            kernel_metadata[kernel_package["version"]] = metadata

        return {
            "versions": kernels_by_version,
            "metadata": kernel_metadata
        }



from typing import TypedDict


class RawSirenadaTestCase(TypedDict):
    test_name: str
    class_name: str
    file_name: str
    browser_type: str
    cluster_type: str
    status: str
    duration: float
    message: str
    start_time: str
    stack_trace: str
    screenshot_file: str
    s3_folder_id: str
    requests_file: str
    sirenada_test_id: str
    sirenada_user: str
    sirenada_password: str


class RawSirenadaRequest(TypedDict):
    build_id: str
    build_job_url: str
    run_id: str # UUID
    region: list[str]
    browsers: list[str]
    clusters: list[str]
    sct_test_id: str
    results: list[RawSirenadaTestCase]


class SirenadaPluginException(Exception):
    pass



import click
import logging
from argus.backend.util.enums import TestStatus

from argus.client.generic.client import ArgusGenericClient

LOGGER = logging.getLogger(__name__)


@click.group
def cli():
    pass


@click.command("submit")
@click.option("--api-key", help="Argus API key for authorization", required=True)
@click.option("--base-url", default="https://argus.scylladb.com", help="Base URL for argus instance")
@click.option("--id", required=True, help="UUID (v4 or v1) unique to the job")
@click.option("--build-id", required=True, help="Unique job identifier in the build system, e.g. scylla-master/group/job for jenkins (The full path)")
@click.option("--build-url", required=True, help="Job URL in the build system")
@click.option("--started-by", required=True, help="Username of the user who started the job")
@click.option("--scylla-version", required=False, default=None, help="Version of Scylla used for this job")
def submit_run(api_key: str, base_url: str, id: str, build_id: str, build_url: str, started_by: str, scylla_version: str = None):
    LOGGER.info("Submitting %s (%s) to Argus...", build_id, id)
    client = ArgusGenericClient(auth_token=api_key, base_url=base_url)
    client.submit_generic_run(build_id=build_id, run_id=id, started_by=started_by, build_url=build_url, scylla_version=scylla_version)
    LOGGER.info("Done.")


@click.command("finish")
@click.option("--api-key", help="Argus API key for authorization", required=True)
@click.option("--base-url", default="https://argus.scylladb.com", help="Base URL for argus instance")
@click.option("--id", required=True, help="UUID (v4 or v1) unique to the job")
@click.option("--status", required=True, help="Resulting job status")
@click.option("--scylla-version", required=False, default=None, help="Version of Scylla used for this job")
def finish_run(api_key: str, base_url: str, id: str, status: str, scylla_version: str = None):
    client = ArgusGenericClient(auth_token=api_key, base_url=base_url)
    status = TestStatus(status)
    client.finalize_generic_run(run_id=id, status=status, scylla_version=scylla_version)


cli.add_command(submit_run)
cli.add_command(finish_run)


if __name__ == "__main__":
    cli()


from flask import Blueprint

from argus.backend.plugins.core import PluginInfoBase, PluginModelBase
from argus.backend.plugins.generic.model import GenericRun


class PluginInfo(PluginInfoBase):
    # pylint: disable=too-few-public-methods
    name: str = "generic"
    model: PluginModelBase = GenericRun
    controller: Blueprint = None
    all_models = [
        GenericRun
    ]
    all_types = [
    ]



from uuid import UUID
from argus.backend.db import ScyllaCluster
from argus.backend.plugins.core import PluginModelBase
from argus.backend.plugins.loader import AVAILABLE_PLUGINS
from argus.backend.util.enums import TestStatus


class ClientException(Exception):
    pass


class ClientService:
    PLUGINS = {name: plugin.model for name, plugin in AVAILABLE_PLUGINS.items()}

    def __init__(self) -> None:
        self.cluster = ScyllaCluster.get()

    def get_model(self, run_type: str) -> PluginModelBase:
        cls = self.PLUGINS.get(run_type)
        if not cls:
            raise ClientException(f"Unsupported run type: {run_type}", run_type)
        return cls

    def submit_run(self, run_type: str, request_data: dict) -> str:
        model = self.get_model(run_type)
        model.submit_run(request_data=request_data)
        return "Created"

    def heartbeat(self, run_type: str, run_id: str) -> int:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        run.update_heartbeat()
        run.save()
        return run.heartbeat

    def get_run_status(self, run_type: str, run_id: str) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        return run.status

    def update_run_status(self, run_type: str, run_id: str, new_status: str) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        run.change_status(new_status=TestStatus(new_status))
        run.save()

        return run.status

    def submit_product_version(self, run_type: str, run_id: str, version: str) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        run.submit_product_version(version)
        run.save()

        return "Submitted"

    def submit_logs(self, run_type: str, run_id: str, logs: list[dict]) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        run.submit_logs(logs)
        run.save()

        return "Submitted"

    def finish_run(self, run_type: str, run_id: str, payload: dict | None = None) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        run.finish_run(payload)
        run.save()

        return "Finalized"



from argus.backend.db import ScyllaCluster
from argus.backend.models.web import ArgusRelease, ArgusTest
from argus.backend.plugins.driver_matrix_tests.model import DriverTestRun


class DriverMatrixService:
    def tested_versions_report(self, build_id: str) -> dict:
        db = ScyllaCluster.get()
        all_runs_for_test_query = db.prepare(f"SELECT * FROM {DriverTestRun.table_name()} WHERE build_id = ?")

        rows = list(db.session.execute(all_runs_for_test_query, parameters=(build_id,)).all())

        if len(rows) == 0:
            raise Exception(f"No results for build_id {build_id}", build_id)

        latest = rows[0]
        try:
            test: ArgusTest = ArgusTest.get(id=latest["test_id"])
            release: ArgusRelease = ArgusRelease.get(id=latest["release_id"])
        except (ArgusTest.DoesNotExist, ArgusRelease.DoesNotExist):
            raise Exception(f"Unable to find release and test information for build_id {build_id} and run_id {latest['id']}", build_id, latest["id"])
        

        version_map = {}

        for row in rows:
            driver_versions = [(col["name"], col["driver"]) for col in row["test_collection"]]
            for version, driver_type in driver_versions:
                driver_type = driver_type if driver_type else "Unknown"
                versions: list = version_map.get(driver_type, [])
                if version not in versions:
                    versions.append(version)
                    versions.sort()
                version_map[driver_type] = versions

        response = {
            "release": release.name,
            "test": test.name,
            "build_id": build_id,
            "versions": version_map,
        }
        return response



from enum import Enum
import logging
from datetime import datetime
from dataclasses import dataclass, field
from typing import Optional
from uuid import UUID

from cassandra.cqlengine import columns
from cassandra.cqlengine.models import _DoesNotExist
from argus.backend.db import ScyllaCluster
from argus.backend.models.web import ArgusRelease
from argus.backend.plugins.core import PluginModelBase
from argus.backend.plugins.sct.resource_setup import ResourceSetup
from argus.backend.plugins.sct.udt import (
    CloudInstanceDetails,
    CloudResource,
    CloudSetupDetails,
    EventsBySeverity,
    NemesisRunInfo,
    PackageVersion,
    PerformanceHDRHistogram
)

LOGGER = logging.getLogger(__name__)
SCT_REGION_PROPERTY_MAP = {
    "aws": "region_name",
    "aws-siren": "region_name",
    "k8s-eks": "region_name",
    "gce": "gce_datacenter",
    "gce-siren": "gce_datacenter",
    "k8s-gke": "gce_datacenter",
    "azure": "azure_region_name",
    "default": "region_name",
}


class SubtestType(str, Enum):
    GEMINI = "gemini"
    PERFORMANCE = "performance"


@dataclass(init=True, repr=True, frozen=True)
class SCTTestRunSubmissionRequest():
    schema_version: str
    run_id: str
    job_name: str
    job_url: str
    started_by: str
    commit_id: str
    sct_config: dict | None
    origin_url: Optional[str] = field(default=None)
    branch_name: Optional[str] = field(default=None)
    runner_public_ip: Optional[str] = field(default=None)
    runner_private_ip: Optional[str] = field(default=None)


class SCTTestRun(PluginModelBase):
    __table_name__ = "sct_test_run"
    _plugin_name = "scylla-cluster-tests"

    # Test Details
    test_name = columns.Text()
    stress_duration = columns.Float()
    scm_revision_id = columns.Text()
    branch_name = columns.Text()
    origin_url = columns.Text()
    started_by = columns.Text()
    config_files = columns.List(value_type=columns.Text())
    packages = columns.List(value_type=columns.UserDefinedType(user_type=PackageVersion))
    scylla_version = columns.Text()
    yaml_test_duration = columns.Integer()

    # Test Preset Resources
    sct_runner_host = columns.UserDefinedType(user_type=CloudInstanceDetails)
    region_name = columns.List(value_type=columns.Text())
    cloud_setup = columns.UserDefinedType(user_type=CloudSetupDetails)

    # Test Runtime Resources
    allocated_resources = columns.List(value_type=columns.UserDefinedType(user_type=CloudResource))

    # Test Results
    events = columns.List(value_type=columns.UserDefinedType(user_type=EventsBySeverity))
    nemesis_data = columns.List(value_type=columns.UserDefinedType(user_type=NemesisRunInfo))
    screenshots = columns.List(value_type=columns.Text())

    # Subtest
    subtest_name = columns.Text()

    # Gemini-related fields
    oracle_nodes_count = columns.Integer()
    oracle_node_ami_id = columns.Text()
    oracle_node_instance_type = columns.Text()
    oracle_node_scylla_version = columns.Text()
    gemini_command = columns.Text()
    gemini_version = columns.Text()
    gemini_status = columns.Text()
    gemini_seed = columns.Text()
    gemini_write_ops = columns.Integer()
    gemini_write_errors = columns.Integer()
    gemini_read_ops = columns.Integer()
    gemini_read_errors = columns.Integer()

    # Performance fields
    perf_op_rate_average = columns.Double()
    perf_op_rate_total = columns.Double()
    perf_avg_latency_99th = columns.Double()
    perf_avg_latency_mean = columns.Double()
    perf_total_errors = columns.Double()
    stress_cmd = columns.Text()

    histograms = columns.List(value_type=columns.Map(key_type=columns.Text(), value_type=columns.UserDefinedType(user_type=PerformanceHDRHistogram)))

    @classmethod
    def _stats_query(cls) -> str:
        return ("SELECT id, test_id, group_id, release_id, status, start_time, build_job_url, build_id, "
                f"assignee, end_time, investigation_status, heartbeat, scylla_version FROM {cls.table_name()} WHERE release_id = ?")

    @classmethod
    def load_test_run(cls, run_id: UUID) -> 'SCTTestRun':
        return cls.get(id=run_id)

    @classmethod
    def submit_run(cls, request_data: dict) -> 'SCTTestRun':
        req = SCTTestRunSubmissionRequest(**request_data)
        return cls.from_sct_config(req=req)

    @classmethod
    def get_distinct_product_versions(cls, release: ArgusRelease) -> list[str]:
        cluster = ScyllaCluster.get()
        statement = cluster.prepare(f"SELECT scylla_version FROM {cls.table_name()} WHERE release_id = ?")
        rows = cluster.session.execute(query=statement, parameters=(release.id,))
        unique_versions = {r["scylla_version"] for r in rows if r["scylla_version"]}

        return sorted(list(unique_versions), reverse=True)

    @classmethod
    def get_version_data_for_release(cls, release_name: str):
        cluster = ScyllaCluster.get()
        release = ArgusRelease.get(name=release_name)
        query = cluster.prepare(f"SELECT scylla_version, packages, status FROM {cls.table_name()} WHERE release_id = ?")
        rows = cluster.session.execute(query=query, parameters=(release.id,))

        return list(rows)

    @classmethod
    def get_perf_results_for_test_name(cls, build_id: str, start_time: float, test_name: str):
        cluster = ScyllaCluster.get()
        query = cluster.prepare(f"SELECT build_id, packages, scylla_version, test_name, perf_op_rate_average, perf_op_rate_total, "
                                "perf_avg_latency_99th, perf_avg_latency_mean, perf_total_errors, id, start_time, build_job_url"
                                f" FROM {cls.table_name()} WHERE build_id = ? AND start_time < ? AND test_name = ? ALLOW FILTERING")
        rows = cluster.session.execute(query=query, parameters=(build_id, start_time, test_name))

        return list(rows)

    @classmethod
    def init_sct_run(cls, req: SCTTestRunSubmissionRequest):
        run = cls()
        run.build_id = req.job_name
        run.assign_categories()
        try:
            run.assignee = run.get_scheduled_assignee()
        except _DoesNotExist:
            run.assignee = None
        run.start_time = datetime.utcnow()
        run.id = UUID(req.run_id)  # pylint: disable=invalid-name
        run.scm_revision_id = req.commit_id
        if req.origin_url:
            run.origin_url = req.origin_url
            run.branch_name = req.branch_name
        run.started_by = req.started_by
        run.build_job_url = req.job_url

        return run

    @classmethod
    def from_sct_config(cls, req: SCTTestRunSubmissionRequest):
        try:
            run = cls.get(id=req.run_id)
        except cls.DoesNotExist:
            run = cls.init_sct_run(req)
            run.save()

        if req.sct_config:
            backend = req.sct_config.get("cluster_backend")
            if duration_override := req.sct_config.get("stress_duration"):
                run.stress_duration = float(duration_override)
            region_key = SCT_REGION_PROPERTY_MAP.get(backend, SCT_REGION_PROPERTY_MAP["default"])
            raw_regions = req.sct_config.get(region_key) or "undefined_region"
            regions = raw_regions.split() if isinstance(raw_regions, str) else raw_regions
            primary_region = regions[0]
            if req.runner_public_ip:  # NOTE: Legacy support, not needed otherwise
                run.sct_runner_host = CloudInstanceDetails(
                    public_ip=req.runner_public_ip,
                    private_ip=req.runner_private_ip,
                    provider=backend,
                    region=primary_region,
                )
            run.cloud_setup = ResourceSetup.get_resource_setup(backend=backend, sct_config=req.sct_config)

            run.config_files = req.sct_config.get("config_files")
            run.region_name = regions
            run.save()

        return run

    def get_resources(self) -> list[CloudResource]:
        return self.allocated_resources

    def get_nemeses(self) -> list[NemesisRunInfo]:
        return self.nemesis_data

    def get_events(self) -> list[EventsBySeverity]:
        return self.events

    def submit_product_version(self, version: str):
        self.scylla_version = version

    def finish_run(self, payload: dict = None):
        self.end_time = datetime.utcnow()

    def submit_logs(self, logs: list[dict]):
        for log in logs:
            self.logs.append((log["log_name"], log["log_link"]))

    def add_screenshot(self, screenshot_link: str):
        self.screenshots.append(screenshot_link)

    def add_nemesis(self, nemesis: NemesisRunInfo):
        self.nemesis_data.append(nemesis)

    def _add_new_event_type(self, event: EventsBySeverity):
        self.events.append(event)

    def _collect_event_message(self, event: EventsBySeverity, message: str):
        if len(event.last_events) >= 100:
            event.last_events = event.last_events[1:]

        event.event_amount += 1
        event.last_events.append(message)

    def add_event(self, event_severity: str, event_message: str):
        try:
            event = next(filter(lambda v: v.severity ==
                         event_severity, self.events))
        except StopIteration:
            event = EventsBySeverity(
                severity=event_severity, event_amount=0, last_events=[])
            self._add_new_event_type(event)

        self._collect_event_message(event, event_message)



from functools import reduce
from datetime import datetime
from typing import TypedDict, Literal
import re
import logging
from uuid import UUID
from pathlib import Path
from pprint import pformat
from glob import glob
from xml.etree import ElementTree

from argus.backend.util.enums import TestStatus
from argus.client.base import ArgusClient
from argus.backend.plugins.driver_matrix_tests.raw_types import RawMatrixTestResult, RawMatrixTestCase


LOGGER = logging.getLogger(__name__)

TestTypeType = Literal['java', 'cpp', 'python', 'gocql']

class AdaptedXUnitData(TypedDict):
    timestamp: str


def python_driver_matrix_adapter(xml: ElementTree.ElementTree) -> AdaptedXUnitData:
    testsuites = list(xml.getroot().iter("testsuite"))

    return {
        "timestamp": testsuites[0].attrib.get("timestamp"),
    }


def java_driver_matrix_adapter(xml: ElementTree.ElementTree) -> AdaptedXUnitData:
    testsuites = xml.getroot()
    ts_now = datetime.utcnow().timestamp()
    try:
        time_taken = float(testsuites.attrib.get("time"))
    except ValueError:
        time_taken = 0.0

    timestamp = datetime.utcfromtimestamp(ts_now - time_taken).isoformat()

    return {
        "timestamp": timestamp,
    }


def cpp_driver_matrix_adapter(xml: ElementTree.ElementTree) -> AdaptedXUnitData:
    testsuites = xml.getroot()

    return {
        "timestamp": testsuites.attrib.get("timestamp"),
    }


def gocql_driver_matrix_adapter(xml: ElementTree.ElementTree) -> AdaptedXUnitData:
    testsuites = list(xml.getroot().iter("testsuite"))

    return {
        "timestamp": testsuites[0].attrib.get("timestamp"),
    }


def generic_adapter(xml: ElementTree.ElementTree) -> AdaptedXUnitData:
    return {
        "timestamp": datetime.utcnow().isoformat()
    }


class ArgusDriverMatrixClient(ArgusClient):
    test_type = "driver-matrix-tests"
    schema_version: None = "v1"

    TEST_ADAPTERS = {
        "java": java_driver_matrix_adapter,
        "cpp": cpp_driver_matrix_adapter,
        "python": python_driver_matrix_adapter,
        "gocql": gocql_driver_matrix_adapter,
    }

    def __init__(self, run_id: UUID, auth_token: str, base_url: str, api_version="v1") -> None:
        super().__init__(auth_token, base_url, api_version)
        self.run_id = run_id

    def parse_build_environment(self, env_path: str) -> dict[str, str]:
        with open(Path(env_path), mode="rt", encoding="utf-8") as env_file:
            raw_env = env_file.read()

        result = {}
        for line in raw_env.split("\n"):
            if not line:
                continue
            LOGGER.debug("ENV: %s", line)
            key, val = line.split(": ")
            result[key] = val.strip()

        return result

    def get_test_cases(self, cases: list[ElementTree.Element]) -> list[RawMatrixTestCase]:
        raw_cases = []
        for case in cases:
            children = list(case.findall("./*"))
            if len(children) > 0:
                status = children[0].tag
                message = f"{children[0].attrib.get('message', 'no-message')} ({children[0].attrib.get('type', 'no-type')})"
            else:
                status = "passed"
                message = ""

            raw_cases.append({
                "name": case.attrib["name"],
                "status": status,
                "time": float(case.attrib["time"]),
                "classname": case.attrib["classname"],
                "message": message,
            })

        return raw_cases

    def get_driver_info(self, xml_name: str, test_type: TestTypeType) -> dict[str, str]:
        if test_type == "cpp":
            filename_re = r"TEST-(?P<driver_name>[\w]*)-(?P<version>[\d\.]*)\.xml"
        else:
            filename_re = r"(?P<name>[\w]*)\.(?P<driver_name>[\w]*)\.(?P<proto>v\d)\.(?P<version>[\d\.]*)\.xml"

        match = re.match(filename_re, xml_name)

        return match.groupdict() if match else {}

    def get_passed_count(self, suite_attribs: dict[str, str]) -> int:
        if (pass_count := suite_attribs.get("passed")):
            return int(pass_count)
        total = int(suite_attribs.get("tests", 0))
        errors = int(suite_attribs.get("errors", 0))
        skipped = int(suite_attribs.get("skipped", 0))
        failures = int(suite_attribs.get("failures", 0))

        return total - errors - skipped - failures

    def parse_result_xml(self, xml_path: Path, test_type:  TestTypeType) -> RawMatrixTestResult:
        with xml_path.open(mode="rt", encoding="utf-8") as xml_file:
            xml = ElementTree.parse(source=xml_file)
            LOGGER.info("%s", pformat(xml))
        testsuites = xml.getroot()
        adapted_data = self.TEST_ADAPTERS.get(test_type, generic_adapter)(xml)

        driver_info = self.get_driver_info(xml_path.name, test_type)
        test_collection = {
            "timestamp": adapted_data["timestamp"],
            "name": xml_path.stem,
            "driver_name": driver_info.get("driver_name"),
            "tests": 0,
            "failures": 0,
            "errors": 0,
            "disabled": 0,
            "skipped": 0,
            "passed": 0,
            "time": 0.0,
        }
        all_suites = []
        for suite in testsuites.iter("testsuite"):
            raw_suite = {
                "name": suite.attrib["name"],
                "tests": int(suite.attrib.get("tests", 0)),
                "failures": int(suite.attrib.get("failures", 0)),
                "disabled": int(0),
                "passed": self.get_passed_count(suite.attrib),
                "skipped": int(suite.attrib.get("skipped", 0)),
                "errors": int(suite.attrib.get("errors", 0)),
                "time": float(suite.attrib["time"]),
                "cases": self.get_test_cases(list(suite.iter("testcase")))
            }
            all_suites.append(raw_suite)
            test_collection["tests"] += raw_suite["tests"]
            test_collection["failures"] += raw_suite["failures"]
            test_collection["errors"] += raw_suite["errors"]
            test_collection["disabled"] += raw_suite["disabled"]
            test_collection["skipped"] += raw_suite["skipped"]
            test_collection["passed"] += raw_suite["passed"]
            test_collection["time"] += raw_suite["time"]

        return {
            **test_collection,
            "suites": all_suites
        }

    def get_results(self, result_path: str, test_type: TestTypeType) -> list[RawMatrixTestResult]:
        xmls = glob(f"{result_path}/**/*.xml", recursive=True)
        LOGGER.info("Will use following XMLs: %s", pformat(xmls))
        results = []
        for xml in xmls:
            result = self.parse_result_xml(Path(xml), test_type)
            results.append(result)
        return results

    def submit(self, build_id: str, build_url: str, env_path: str, result_path: str, test_type: TestTypeType):
        env = self.parse_build_environment(env_path)
        results = self.get_results(result_path, test_type)

        self.submit_driver_matrix_run(job_name=build_id, job_url=build_url, test_environment=env, results=results)

    def submit_driver_matrix_run(self, job_name: str, job_url: str,
                                 results: list[RawMatrixTestResult], test_environment: dict[str, str]) -> None:
        response = super().submit_run(run_type=self.test_type, run_body={
            "run_id": str(self.run_id),
            "job_name": job_name,
            "job_url": job_url,
            "test_environment": test_environment,
            "matrix_results": results
        })

        self.check_response(response)

    def set_matrix_status(self, status: TestStatus):
        response = super().set_status(run_type=self.test_type, run_id=self.run_id, new_status=status)
        self.check_response(response)



from cassandra.cqlengine.usertype import UserType
from cassandra.cqlengine import columns


class TestCase(UserType):
    name = columns.Text()
    status = columns.Text()
    time = columns.Float()
    classname = columns.Text()
    message = columns.Text()


class TestSuite(UserType):
    name = columns.Text()
    tests_total = columns.Integer()
    failures = columns.Integer()
    disabled = columns.Integer()
    skipped = columns.Integer()
    passed = columns.Integer()
    errors = columns.Integer()
    time = columns.Float()
    cases = columns.List(value_type=columns.UserDefinedType(user_type=TestCase))


class TestCollection(UserType):
    name = columns.Text()
    driver = columns.Text()
    tests_total = columns.Integer()
    failures = columns.Integer()
    disabled = columns.Integer()
    skipped = columns.Integer()
    passed = columns.Integer()
    errors = columns.Integer()
    timestamp = columns.DateTime()
    time = columns.Float()
    suites = columns.List(value_type=columns.UserDefinedType(user_type=TestSuite))


class EnvironmentInfo(UserType):
    key = columns.Text()
    value = columns.Text()



import subprocess
import json
import logging
import datetime
from types import NoneType
from uuid import UUID
from cassandra.util import uuid_from_time  # pylint: disable=no-name-in-module
from flask import current_app
from argus.backend.db import ScyllaCluster
from argus.backend.plugins.loader import AVAILABLE_PLUGINS, all_plugin_models
from argus.backend.plugins.sct.testrun import SCTTestRun
from argus.backend.service.notification_manager import NotificationManagerService
from argus.backend.models.web import (
    ArgusRelease,
    ArgusGroup,
    ArgusTest,
    ArgusSchedule,
    ArgusScheduleAssignee,
    ArgusScheduleGroup,
    ArgusScheduleTest,
    ArgusTestRunComment,
    ArgusEvent,
    ReleasePlannerComment,
    User,
)
from argus.backend.events.event_processors import EVENT_PROCESSORS
from argus.backend.service.testrun import TestRunService

LOGGER = logging.getLogger(__name__)


class ArgusService:
    # pylint: disable=no-self-use,too-many-arguments,too-many-instance-attributes,too-many-locals, too-many-public-methods
    def __init__(self, database_session=None):
        self.session = database_session if database_session else ScyllaCluster.get_session()
        self.database = ScyllaCluster.get()
        self.notification_manager = NotificationManagerService()
        self.github_headers = {
            "Accept": "application/vnd.github.v3+json",
            "Authorization": f"token {current_app.config['GITHUB_ACCESS_TOKEN']}"
        }
        self.build_id_and_url_statement = self.database.prepare(
            f"SELECT build_id, build_job_url, test_id FROM {SCTTestRun.table_name()} WHERE id = ?"
        )  # TODO: transfer to PluginModelBase
        self.scylla_versions_by_release = self.database.prepare(
            f"SELECT scylla_version FROM {SCTTestRun.table_name()} WHERE release_id = ?"
        )  # TODO: Moved to PluginModelBase

    def get_version(self) -> str:
        try:
            proc = subprocess.run(["git", "rev-parse", "HEAD"], check=True, capture_output=True)
        except subprocess.CalledProcessError:
            proc = None
        if proc:
            return proc.stdout.decode(encoding="utf-8").strip()
        else:
            try:
                with open("./.argus_version", 'rt', encoding="utf-8") as version_file:
                    version = version_file.read().strip()
                return version
            except FileNotFoundError:
                return "version_unknown"

    def create_release(self, payload: dict) -> dict:
        response = {}
        for release_name in payload:
            try:
                ArgusRelease.get(name=release_name)
                response[release_name] = {
                    "status": "error",
                    "message": f"Release {release_name} already exists"
                }
                continue
            except ArgusRelease.DoesNotExist:
                pass

            new_release = ArgusRelease()
            new_release.name = release_name
            new_release.save()
            response[release_name] = {}
            response[release_name]["groups"] = self.create_groups(
                groups=payload[release_name]["groups"],
                parent_release_id=new_release.id
            )

        return response

    def create_groups(self, groups: dict, parent_release_id) -> dict:
        response = {}
        for group_name, group_definition in groups.items():
            new_group = ArgusGroup()
            new_group.release_id = parent_release_id
            new_group.name = group_name
            new_group.pretty_name = group_definition.get("pretty_name")
            new_group.save()
            response[group_name] = {}
            response[group_name]["status"] = "created"
            response[group_name]["tests"] = self.create_tests(
                tests=group_definition.get("tests", []),
                parent_group_id=new_group.id,
                parent_release_id=parent_release_id
            )
        return response

    def create_tests(self, tests: dict, parent_group_id: UUID, parent_release_id: UUID) -> dict:
        response = {}

        for test_name in tests:
            new_test = ArgusTest()
            new_test.release_id = parent_release_id
            new_test.group_id = parent_group_id
            new_test.name = test_name
            new_test.save()
            response[test_name] = "created"

        return response

    def get_comment(self, comment_id: UUID) -> ArgusTestRunComment | None:
        try:
            return ArgusTestRunComment.get(id=comment_id)
        except ArgusTestRunComment.DoesNotExist:
            return None

    def get_releases(self):
        releases = list(ArgusRelease.all())
        releases = sorted(releases, key=lambda r: r.name)
        releases = sorted(releases, key=lambda r: r.dormant)
        return releases

    def get_groups(self, release_id: UUID) -> list[ArgusGroup]:
        groups = list(ArgusGroup.filter(release_id=release_id).all())
        return sorted(groups, key=lambda g: g.pretty_name if g.pretty_name else g.name)

    def get_groups_for_release(self, release: ArgusRelease):
        groups = ArgusGroup.filter(release_id=release.id).all()
        return sorted(groups, key=lambda g: g.pretty_name if g.pretty_name else g.name)

    def get_tests(self, group_id: UUID) -> list[ArgusTest]:
        return list(ArgusTest.filter(group_id=group_id).all())

    def get_test_info(self, test_id: UUID) -> dict:
        test = ArgusTest.get(id=test_id)
        group = ArgusGroup.get(id=test.group_id)
        release = ArgusRelease.get(id=test.release_id)
        return {
            "test": dict(test.items()),
            "group": dict(group.items()),
            "release": dict(release.items()),
        }

    def get_data_for_release_dashboard(self, release_name: str):
        release = ArgusRelease.get(name=release_name)
        release_groups = ArgusGroup.filter(release_id=release.id).all()
        release_tests = ArgusTest.filter(release_id=release.id).all()

        return release, release_groups, release_tests

    def get_distinct_release_versions(self, release_id: UUID | str) -> list[str]:
        release_id = UUID(release_id) if isinstance(release_id, str) else release_id
        release = ArgusRelease.get(id=release_id)
        unique_versions = {ver for plugin in all_plugin_models()
                           for ver in plugin.get_distinct_product_versions(release=release)}

        return sorted(list(unique_versions), reverse=True)

    def poll_test_runs(self, test_id: UUID, additional_runs: list[UUID], limit: int = 10):
        test: ArgusTest = ArgusTest.get(id=test_id)

        rows: list[SCTTestRun] = list(SCTTestRun.filter(build_id=test.build_system_id).all().limit(limit))

        rows_ids = [row.id for row in rows]

        for run_id in additional_runs:
            if run_id not in rows_ids:
                row: SCTTestRun = SCTTestRun.get(id=run_id)
                rows.append(row)

        for row in rows:
            try:
                setattr(row, "build_number", int(row["build_job_url"].rstrip("/").split("/")[-1]))
            except ValueError:
                setattr(row, "build_number", -1)

        return sorted(rows, reverse=True, key=lambda r: r.build_number)

    def poll_test_runs_single(self, runs: list[UUID]):
        rows: list[SCTTestRun] = []
        for run_id in runs:
            try:
                row: SCTTestRun = SCTTestRun.get(id=run_id)
                rows.append(row)
            except SCTTestRun.DoesNotExist:
                pass

        response = {str(row.id): row for row in rows}
        return response

    def fetch_release_activity(self, release_name: str) -> dict:
        response = {}
        release = ArgusRelease.get(name=release_name)
        all_events = ArgusEvent.filter(release_id=release.id).all()
        all_events = sorted(all_events, key=lambda ev: ev.created_at)
        response["release_id"] = release.id
        response["raw_events"] = [dict(event.items()) for event in all_events]
        response["events"] = {str(event.id): EVENT_PROCESSORS.get(
            event.kind)(json.loads(event.body)) for event in all_events}
        return response

    def assign_runs_for_scheduled_test(self, schedule: ArgusSchedule, test_id: UUID, new_assignee: UUID):
        test: ArgusTest = ArgusTest.get(id=test_id)
        affected_rows: list[SCTTestRun] = list(SCTTestRun.filter(
            build_id=test.build_system_id,
            start_time__gte=schedule.period_start,
            start_time__lte=schedule.period_end
        ).all()
        )
        for row in affected_rows:
            if row.assignee != new_assignee:
                row.assignee = new_assignee
                row.save()

    def assign_runs_for_scheduled_group(self, schedule: ArgusSchedule, group_id: UUID, new_assignee: UUID):
        tests = ArgusTest.filter(group_id=group_id).all()
        build_ids = [test.build_system_id for test in tests]
        affected_rows: list[SCTTestRun] = list(SCTTestRun.filter(
            start_time__gte=schedule.period_start,
            start_time__lte=schedule.period_end
        ).all()
        )
        for row in affected_rows:
            if row.build_id in build_ids and row.assignee != new_assignee:
                row.assignee = new_assignee
                row.save()

    def submit_new_schedule(self, release: str | UUID, start_time: str, end_time: str, tests: list[str | UUID],
                            groups: list[str | UUID], assignees: list[str | UUID], tag: str) -> dict:
        release = UUID(release) if isinstance(release, str) else release
        if len(assignees) == 0:
            raise Exception("Assignees not specified in the new schedule")

        if len(tests) == 0 and len(groups) == 0:
            raise Exception("Schedule does not contain scheduled objects")

        schedule = ArgusSchedule()
        schedule.release_id = release
        schedule.period_start = datetime.datetime.fromisoformat(start_time)
        schedule.period_end = datetime.datetime.fromisoformat(end_time)
        schedule.id = uuid_from_time(schedule.period_start)
        schedule.tag = tag
        schedule.save()

        response = dict(schedule.items())
        response["assignees"] = []
        response["tests"] = []
        response["groups"] = []

        for test_id in tests:
            test_entity = ArgusScheduleTest()
            test_entity.id = uuid_from_time(schedule.period_start)
            test_entity.schedule_id = schedule.id
            test_entity.test_id = UUID(test_id) if isinstance(test_id, str) else test_id
            test_entity.release_id = release
            test_entity.save()
            self.assign_runs_for_scheduled_test(schedule, test_entity.test_id, UUID(assignees[0]))
            response["tests"].append(test_id)

        for group_id in groups:
            group_entity = ArgusScheduleGroup()
            group_entity.id = uuid_from_time(schedule.period_start)
            group_entity.schedule_id = schedule.id
            group_entity.group_id = UUID(group_id) if isinstance(group_id, str) else group_id
            group_entity.release_id = release
            group_entity.save()
            self.assign_runs_for_scheduled_group(schedule, group_entity.group_id, UUID(assignees[0]))
            response["groups"].append(group_id)

        for assignee_id in assignees:
            assignee_entity = ArgusScheduleAssignee()
            assignee_entity.id = uuid_from_time(schedule.period_start)
            assignee_entity.schedule_id = schedule.id
            assignee_entity.assignee = UUID(assignee_id) if isinstance(assignee_id, str) else assignee_id
            assignee_entity.release_id = release
            assignee_entity.save()
            response["assignees"].append(assignee_id)

        return response

    def get_schedules_for_release(self, release_id: str | UUID) -> dict:
        """
        {
            "release": "hex-uuid"
        }
        """
        release_id = UUID(release_id) if isinstance(release_id, str) else release_id
        release: ArgusRelease = ArgusRelease.get(id=release_id)
        if release.perpetual:
            today = datetime.datetime.utcnow()
            six_months_ago = today - datetime.timedelta(days=180)
            uuid_six_months = uuid_from_time(six_months_ago)
            schedules = ArgusSchedule.filter(release_id=release_id, id__gte=uuid_six_months).all()
        else:
            schedules = ArgusSchedule.filter(release_id=release_id).all()
        response = {
            "schedules": []
        }
        for schedule in schedules:
            serialized_schedule = dict(schedule.items())
            tests = ArgusScheduleTest.filter(schedule_id=schedule.id).all()
            serialized_schedule["tests"] = [test.test_id for test in tests]
            groups = ArgusScheduleGroup.filter(schedule_id=schedule.id).all()
            serialized_schedule["groups"] = [group.group_id for group in groups]
            assignees = ArgusScheduleAssignee.filter(schedule_id=schedule.id).all()
            serialized_schedule["assignees"] = [assignee.assignee for assignee in assignees]
            response["schedules"].append(serialized_schedule)

        return response

    def update_schedule_assignees(self, payload: dict) -> dict:
        """
        {
            "releaseId": "hex-uuid",
            "scheduleId": "hex-uuid",
        }
        """
        release_id = payload.get("releaseId")
        if not release_id:
            raise Exception("Release name not specified in the request")

        schedule_id = payload.get("scheduleId")
        if not schedule_id:
            raise Exception("No schedule Id provided")

        assignees = payload.get("newAssignees")
        if not assignees:
            raise Exception("No assignees provided")

        release = ArgusRelease.get(id=release_id)
        schedule = ArgusSchedule.get(release_id=release.id, id=schedule_id)
        schedule_tests = ArgusScheduleTest.filter(schedule_id=schedule.id).all()
        schedule_groups = ArgusScheduleGroup.filter(schedule_id=schedule.id).all()
        for test in schedule_tests:
            self.assign_runs_for_scheduled_test(schedule, test.test_id, UUID(assignees[0]))

        for group in schedule_groups:
            self.assign_runs_for_scheduled_group(schedule, group.group_id, UUID(assignees[0]))

        old_assignees = list(ArgusScheduleAssignee.filter(schedule_id=schedule.id).all())
        for new_assignee in assignees:
            assignee = ArgusScheduleAssignee()
            assignee.release_id = release.id
            assignee.schedule_id = schedule.id
            assignee.assignee = UUID(new_assignee)
            assignee.save()

        for old_assignee in old_assignees:
            old_assignee.delete()

        response = {
            "scheduleId": schedule_id,
            "status": "changed assignees",
            "newAssignees": assignees,
        }

        return response

    def update_schedule_comment(self, payload: dict) -> dict:
        new_comment = payload.get("newComment")
        release_id = payload.get("releaseId")
        group_id = payload.get("groupId")
        test_id = payload.get("testId")

        if not release_id:
            raise Exception("No release provided")
        if not group_id:
            raise Exception("No group provided")
        if not test_id:
            raise Exception("No test provided")

        if isinstance(new_comment, NoneType):
            raise Exception("No comment provided in the body of request")

        try:
            comment = ReleasePlannerComment.get(release=release_id, group=group_id, test=test_id)
        except ReleasePlannerComment.DoesNotExist:
            comment = ReleasePlannerComment()
            comment.release = release_id
            comment.group = group_id
            comment.test = test_id

        comment.comment = new_comment
        comment.save()

        return {
            "releaseId": release_id,
            "groupId": group_id,
            "testId": test_id,
            "newComment": new_comment,
        }

    def delete_schedule(self, payload: dict) -> dict:
        """
        {
            "release": hex-uuid,
            "schedule_id": uuid1
        }
        """
        release_id = payload.get("releaseId")
        if not release_id:
            raise Exception("Release name not specified in the request")

        schedule_id = payload.get("scheduleId")
        if not schedule_id:
            raise Exception("Schedule id not specified in the request")

        release = ArgusRelease.get(id=release_id)
        schedule = ArgusSchedule.get(release_id=release.id, id=schedule_id)
        tests = ArgusScheduleTest.filter(schedule_id=schedule.id).all()
        groups = ArgusScheduleGroup.filter(schedule_id=schedule.id).all()
        assignees = ArgusScheduleAssignee.filter(schedule_id=schedule.id).all()

        full_schedule = dict(schedule)
        full_schedule["tests"] = [test.test_id for test in tests]
        full_schedule["groups"] = [group.group_id for group in groups]
        full_schedule["assignees"] = [assignee.assignee for assignee in assignees]

        if len(assignees) > 0:
            schedule_user = User.get(id=assignees[0].assignee)
            service = TestRunService()

            for model in all_plugin_models():
                for run in model.get_jobs_assigned_to_user(schedule_user):
                    if run["release_id"] != release.id:
                        continue
                    if run["test_id"] not in full_schedule["tests"]:
                        continue
                    if schedule.period_start < run["start_time"] < schedule.period_end:
                        service.change_run_assignee(test_id=run["test_id"], run_id=run["id"], new_assignee=None)

        for entities in [tests, groups, assignees]:
            for entity in entities:
                entity.delete()

        schedule.delete()
        return {
            "releaseId": release.id,
            "scheduleId": schedule_id,
            "result": "deleted"
        }

    def get_planner_data(self, release_id: UUID | str) -> dict:

        release = ArgusRelease.get(id=release_id)
        release_comments = list(ReleasePlannerComment.filter(release=release.id).all())
        groups = ArgusGroup.filter(release_id=release.id).all()
        groups_by_group_id = {str(group.id): dict(group.items()) for group in groups if group.enabled}
        tests = ArgusTest.filter(release_id=release.id).all()
        tests = [dict(t.items()) for t in tests if t.enabled]
        tests_by_group = {}
        for test in tests:
            group = groups_by_group_id.get(str(test["group_id"]))
            if not group:
                continue
            test["group_name"] = group["name"]
            test["pretty_group_name"] = groups_by_group_id[str(test["group_id"])]["pretty_name"]
            try:
                comment = next(filter(lambda c: c.test == test["id"], release_comments))
            except StopIteration:
                comment = None
            test["comment"] = comment.comment if comment else ""
            group_tests = tests_by_group.get(test["group_name"], [])
            group_tests.append(test)
            tests_by_group[test["group_name"]] = group_tests

        response = {
            "release": dict(release.items()),
            "groups": groups_by_group_id,
            "tests": tests,
            "tests_by_group": tests_by_group,
        }

        return response

    def get_groups_assignees(self, release_id: UUID | str):
        release_id = UUID(release_id) if isinstance(release_id, str) else release_id
        release = ArgusRelease.get(id=release_id)

        groups = ArgusGroup.filter(release_id=release_id).all()
        group_ids = [group.id for group in groups if group.enabled]

        scheduled_groups = ArgusScheduleGroup.filter(release_id=release.id, group_id__in=group_ids).all()
        schedule_ids = {schedule.schedule_id for schedule in scheduled_groups}

        schedules = ArgusSchedule.filter(release_id=release.id, id__in=tuple(schedule_ids)).all()

        valid_schedules = schedules
        if release.perpetual:
            today = datetime.datetime.utcnow()
            valid_schedules = [s for s in schedules if s.period_start <= today <= s.period_end]

        response = {}
        for schedule in valid_schedules:
            assignees = ArgusScheduleAssignee.filter(schedule_id=schedule.id).all()
            assignees_uuids = [assignee.assignee for assignee in assignees]
            schedule_groups = filter(lambda g: g.schedule_id == schedule.id, scheduled_groups)
            groups = {str(group.group_id): assignees_uuids for group in schedule_groups}
            response = {**groups, **response}

        return response

    def get_tests_assignees(self, group_id: UUID | str):
        group_id = UUID(group_id) if isinstance(group_id, str) else group_id
        group = ArgusGroup.get(id=group_id)

        release = ArgusRelease.get(id=group.release_id)
        tests = ArgusTest.filter(group_id=group_id).all()

        test_ids = [test.id for test in tests if test.enabled]

        scheduled_tests = ArgusScheduleTest.filter(release_id=release.id, test_id__in=tuple(test_ids)).all()
        schedule_ids = {test.schedule_id for test in scheduled_tests}
        schedules: list[ArgusSchedule] = list(ArgusSchedule.filter(
            release_id=release.id, id__in=tuple(schedule_ids)).all())

        if release.perpetual:
            today = datetime.datetime.utcnow()
            schedules = list(filter(lambda s: s.period_start <= today <= s.period_end, schedules))

        response = {}
        for schedule in schedules:
            assignees = ArgusScheduleAssignee.filter(schedule_id=schedule.id).all()
            assignees_uuids = [assignee.assignee for assignee in assignees]
            schedule_tests = filter(lambda t: t.schedule_id == schedule.id, scheduled_tests)
            tests = {str(test.test_id): assignees_uuids for test in schedule_tests}
            response = {**tests, **response}

        return response

    def get_jobs_for_user(self, user: User):
        today = datetime.datetime.now()
        validity_period = today - datetime.timedelta(days=current_app.config.get("JOB_VALIDITY_PERIOD_DAYS", 30))
        for plugin in all_plugin_models():
            for run in plugin.get_jobs_assigned_to_user(user=user):
                if run["start_time"] >= validity_period:
                    yield run

    def get_schedules_for_user(self, user: User) -> list[dict]:
        all_assigned_schedules = ArgusScheduleAssignee.filter(assignee=user.id).all()
        schedule_keys = [(schedule_assignee.release_id, schedule_assignee.schedule_id)
                         for schedule_assignee in all_assigned_schedules]
        schedules = []
        today = datetime.datetime.utcnow()
        for release_id, schedule_id in schedule_keys:
            try:
                schedule = dict(ArgusSchedule.get(release_id=release_id, id=schedule_id).items())
            except ArgusSchedule.DoesNotExist:
                continue
            if schedule["period_start"] <= today <= schedule["period_end"]:
                tests = ArgusScheduleTest.filter(schedule_id=schedule_id).all()
                schedule["tests"] = [test.test_id for test in tests]
                groups = ArgusScheduleGroup.filter(schedule_id=schedule_id).all()
                schedule["groups"] = [group.group_id for group in groups]
                assignees = ArgusScheduleAssignee.filter(schedule_id=schedule_id).all()
                schedule["assignees"] = [assignee.assignee for assignee in assignees]
                schedules.append(schedule)

        return schedules

    def get_planner_comment_by_test(self, test_id):
        try:
            test = ArgusTest.get(id=test_id)
            return ReleasePlannerComment.get(test=test.id, release=test.release_id, group=test.group_id).comment
        except ReleasePlannerComment.DoesNotExist:
            return ""



from typing import TypeVar

T = TypeVar("T")


def is_list_homogeneous(list_to_check: list[T]) -> bool:
    if not len(list_to_check):
        return True

    first, *_ = list_to_check
    first_type = type(first)

    filtered_list = list(filter(lambda val: type(val) != first_type, list_to_check))

    return len(filtered_list) == 0



from flask import Blueprint

from argus.backend.plugins.core import PluginInfoBase, PluginModelBase
from argus.backend.plugins.sirenada.model import SirenadaRun, SirenadaTest


class PluginInfo(PluginInfoBase):
    # pylint: disable=too-few-public-methods
    name: str = "sirenada"
    model: PluginModelBase = SirenadaRun
    controller: Blueprint = None
    all_models = [
        SirenadaRun
    ]
    all_types = [
        SirenadaTest
    ]



from enum import Enum


class NemesisStatus(str, Enum):
    STARTED = "started"
    RUNNING = "running"
    FAILED = "failed"
    SKIPPED = "skipped"
    SUCCEEDED = "succeeded"
    TERMINATED = "terminated"


class TestStatus(str, Enum):
    CREATED = "created"
    RUNNING = "running"
    FAILED = "failed"
    TEST_ERROR = "test_error"
    ERROR = "error"
    PASSED = "passed"
    ABORTED = "aborted"
    NOT_PLANNED = "not_planned"
    NOT_RUN = "not_run"


class TestInvestigationStatus(str, Enum):
    NOT_INVESTIGATED = "not_investigated"
    IN_PROGRESS = "in_progress"
    INVESTIGATED = "investigated"
    IGNORED = "ignored"


class ResourceState(str, Enum):
    RUNNING = "running"
    STOPPED = "stopped"
    TERMINATED = "terminated"



import logging
from uuid import UUID
from flask import (
    Blueprint,
    request
)
from argus.backend.error_handlers import handle_api_exception
from argus.backend.models.web import User
from argus.backend.service.argus_service import ArgusService
from argus.backend.service.user import api_login_required
from argus.backend.util.common import get_payload
from argus.backend.service.team_manager_service import TeamManagerService

bp = Blueprint('team_api', __name__, url_prefix='/team')
LOGGER = logging.getLogger(__name__)
bp.register_error_handler(Exception, handle_api_exception)


@bp.route("/create", methods=["POST"])
@api_login_required
def team_create():
    payload = get_payload(request)
    result = TeamManagerService().create_team(
        name=payload["name"], 
        leader=UUID(payload["leader"]),
        members=[UUID(m) for m in payload["members"]],
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:team_id>/get")
@api_login_required
def team_get(team_id: str):
    result = TeamManagerService().get_team_by_id(UUID(team_id))

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:team_id>/delete", methods=["DELETE"])
@api_login_required
def team_delete(team_id: str):
    TeamManagerService().delete_team(UUID(team_id))

    return {
        "status": "ok",
        "response": {
            "team_id": team_id,
            "status": "deleted",
        }
    }

@bp.route("/<string:team_id>/edit", methods=["POST"])
@api_login_required
def team_edit(team_id: str):
    payload = get_payload(request)
    team = TeamManagerService().edit_team(
        team_id=UUID(payload["id"]),
        name=payload["name"],
        members=[UUID(m) for m in payload["members"]],
    )

    return {
        "status": "ok",
        "response": {
            "team_id": team_id,
            "status": "updated",
            "team": team,
        }
    }

@bp.route("/<string:team_id>/motd/edit", methods=["POST"])
@api_login_required
def team_edit_motd(team_id: str):
    payload = get_payload(request)
    TeamManagerService().edit_team_motd(
        team_id=UUID(payload["id"]),
        message=payload["motd"],
    )

    return {
        "status": "ok",
        "response": {
            "team_id": team_id,
            "status": "updated",
        }
    }

@bp.route("/user/<string:user_id>/teams")
@api_login_required
def user_teams(user_id: str):
    result = TeamManagerService().get_users_teams(user_id=UUID(user_id))

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/user/<string:user_id>/jobs")
@api_login_required
def user_jobs(user_id: str):
    user = User.get(id=UUID(user_id))
    result = list(ArgusService().get_jobs_for_user(user))

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/leader/<string:user_id>/teams")
@api_login_required
def leader_teams(user_id: str):
    result = TeamManagerService().get_teams_for_user(user_id=UUID(user_id))

    return {
        "status": "ok",
        "response": result
    }



import logging
from typing import TypedDict
from uuid import UUID
from argus.backend.db import ScyllaCluster
from argus.backend.plugins.sct.testrun import SCTTestRun
from argus.backend.models.web import ArgusRelease, ArgusGroup, ArgusTest

LOGGER = logging.getLogger(__name__)


class ReleaseManagerException(Exception):
    pass


class ReleaseEditPayload(TypedDict):
    id: str
    pretty_name: str
    description: str
    valid_version_regex: str | None
    enabled: bool
    perpetual: bool
    dormant: bool


class ReleaseManagerService:
    def __init__(self) -> None:
        self.session = ScyllaCluster.get_session()
        self.database = ScyllaCluster.get()
        self.runs_by_build_id_stmt = self.database.prepare(
            "SELECT id, test_id, group_id, release_id, build_id, start_time "
            f"FROM {SCTTestRun.table_name()} WHERE build_id = ?"
        )
        self.update_run_stmt = self.database.prepare(
            f"UPDATE {SCTTestRun.table_name()} SET test_id = ?, group_id = ?, release_id = ? "
            "WHERE build_id = ? AND start_time = ?"
        )

    def get_releases(self) -> list[ArgusRelease]:
        return list(ArgusRelease.all())

    def get_groups(self, release_id: UUID) -> list[ArgusGroup]:
        return list(ArgusGroup.filter(release_id=release_id).all())

    def get_tests(self, group_id: UUID) -> list[ArgusTest]:
        return list(ArgusTest.filter(group_id=group_id).all())

    def create_release(self, release_name: str, pretty_name: str, perpetual: bool) -> ArgusRelease:
        try:
            release = ArgusRelease.get(name=release_name)
        except ArgusRelease.DoesNotExist:
            release = ArgusRelease()
            release.name = release_name
            release.pretty_name = pretty_name
            release.perpetual = perpetual

            release.save()
        else:
            raise ReleaseManagerException(
                f"Release {release_name} already exists!", release_name)

        return release

    def create_group(self, group_name: str, pretty_name: str, build_system_id: str,
                     release_id: str) -> ArgusGroup:
        release = ArgusRelease.get(id=UUID(release_id))

        new_group = ArgusGroup()
        new_group.name = group_name
        new_group.pretty_name = pretty_name
        new_group.release_id = release.id
        new_group.build_system_id = build_system_id
        new_group.save()

        return new_group

    def create_test(self, test_name, pretty_name, build_id, build_url, group_id, release_id, plugin_name) -> ArgusTest:
        release = ArgusRelease.get(id=UUID(release_id))
        group = ArgusGroup.get(id=UUID(group_id))

        new_test = ArgusTest()
        new_test.name = test_name
        new_test.pretty_name = pretty_name
        new_test.build_system_id = build_id
        new_test.release_id = release.id
        new_test.group_id = group.id
        new_test.plugin_name = plugin_name
        new_test.build_system_url = build_url
        new_test.validate_build_system_id()
        new_test.save()
        self.move_test_runs(new_test)
        return new_test

    def delete_group(self, group_id: str, delete_tests: bool = True, new_group_id: str = "") -> bool:
        group_to_delete = ArgusGroup.get(id=UUID(group_id))

        tests_to_change = ArgusTest.filter(
            group_id=group_to_delete.id)
        if delete_tests:
            for test in tests_to_change.all():
                test.delete()
        else:
            new_group = ArgusGroup.get(id=UUID(new_group_id))
            for test in tests_to_change.all():
                test.group_id = new_group.id
                test.save()

        group_to_delete.delete()

        return True

    def delete_test(self, test_id: str) -> bool:
        test_to_delete = ArgusTest.get(id=test_id)
        test_to_delete.delete()

        return True

    def update_group(self, group_id: str, name: str, pretty_name: str, enabled: bool, build_system_id: str) -> bool:
        group = ArgusGroup.get(id=UUID(group_id))

        group.name = name
        group.build_system_id = build_system_id
        group.pretty_name = pretty_name
        group.enabled = enabled

        group.save()

        return True

    def update_test(self, test_id: str, name: str, pretty_name: str, plugin_name: str,
                    enabled: bool, build_system_id: str, build_system_url: str, group_id) -> bool:
        test: ArgusTest = ArgusTest.get(id=UUID(test_id))
        group = ArgusGroup.get(id=UUID(group_id))

        test.name = name
        test.pretty_name = pretty_name
        test.plugin_name = plugin_name
        test.enabled = enabled
        test.build_system_id = build_system_id
        test.build_system_url = build_system_url
        if test.group_id != group.id:
            test.group_id = group.id
            LOGGER.info("Relocating old test runs into a new group")

        test.validate_build_system_id()
        test.save()
        self.move_test_runs(test)
        return True

    def set_release_state(self, release_id: str, state: bool) -> bool:
        release = ArgusRelease.get(id=UUID(release_id))
        release.enabled = state
        release.save()

        return True

    def set_release_dormancy(self, release_id: str, dormant: bool) -> bool:
        release = ArgusRelease.get(id=UUID(release_id))
        release.dormant = dormant
        release.save()

        return True

    def set_release_perpetuality(self, release_id: str, perpetual: bool) -> bool:
        release = ArgusRelease.get(id=UUID(release_id))
        release.perpetual = perpetual
        release.save()

        return True

    def edit_release(self, payload: ReleaseEditPayload) -> bool:

        release: ArgusRelease = ArgusRelease.get(id=payload["id"])
        release.pretty_name = payload["pretty_name"]
        release.perpetual = payload["perpetual"]
        release.enabled = payload["enabled"]
        release.dormant = payload["dormant"]
        release.description = payload["description"]
        release.valid_version_regex = payload["valid_version_regex"]


        release.save()
        return True

    def delete_release(self, release_id: str) -> bool:

        release: ArgusRelease = ArgusRelease.get(id=release_id)

        release_groups = ArgusGroup.filter(release_id=release.id)
        release_tests = ArgusTest.filter(release_id=release.id)

        for entity in [*release_groups.all(), *release_tests.all()]:
            entity.delete()

        release.delete()
        return True

    def batch_move_tests(self, new_group_id: str, tests: list[str]) -> bool:
        group = ArgusGroup.get(id=UUID(new_group_id))

        tests: list[ArgusTest] = [ArgusTest.get(id=UUID(test_id)) for test_id in tests]

        for test in tests:
            test.group_id = group.id
            test.save()
            self.move_test_runs(test)

        return True

    def move_test_runs(self, test: ArgusTest) -> None:
        run_rows = self.session.execute(self.runs_by_build_id_stmt, parameters=(
            test.build_system_id,), execution_profile="read_fast")
        for run in run_rows:
            run["test_id"] = test.id
            run["group_id"] = test.group_id
            run["release_id"] = test.release_id
            self.session.execute(self.update_run_stmt, parameters=run)



import logging

from datetime import datetime
from typing import TypedDict
from uuid import UUID

from argus.backend.plugins.loader import all_plugin_models
from argus.backend.util.common import get_build_number
from argus.backend.util.enums import TestStatus, TestInvestigationStatus
from argus.backend.models.web import ArgusGithubIssue, ArgusRelease, ArgusGroup, ArgusTest,\
    ArgusScheduleTest, ArgusTestRunComment
from argus.backend.db import ScyllaCluster

LOGGER = logging.getLogger(__name__)


class TestRunStatRow(TypedDict):
    build_id: str
    status: TestStatus
    investigation_status: TestInvestigationStatus
    assignee: UUID
    scylla_version: str  # TODO: rework SCT specific field
    start_time: datetime
    end_time: datetime
    heartbeat: int
    id: UUID
    test_id: UUID
    group_id: UUID
    release_id: UUID
    build_job_url: str


class ComparableTestStatus:
    PRIORITY_MAP = {
        TestStatus.FAILED: 10,
        TestStatus.TEST_ERROR: 10,
        TestStatus.ABORTED: 9,
        TestStatus.RUNNING: 8,
        TestStatus.CREATED: 7,
        TestStatus.PASSED: 6,
    }

    def __init__(self, status: TestStatus):
        self._status = status

    def _get_prio(self):
        return self.PRIORITY_MAP.get(self._status, 0)

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestStatus):
            return False
        return self._get_prio() == __o._get_prio()

    def __ne__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestStatus):
            return False
        return not self.__eq__(__o)

    def __lt__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestStatus):
            return False
        return self._get_prio() < __o._get_prio()

    def __gt__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestStatus):
            return False
        return self._get_prio() > __o._get_prio()

    def __ge__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestStatus):
            return False
        return self._get_prio() >= __o._get_prio()

    def __le__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestStatus):
            return False
        return self._get_prio() <= __o._get_prio()


class ComparableTestInvestigationStatus:
    PRIORITY_MAP = {
        TestInvestigationStatus.NOT_INVESTIGATED: 10,
        TestInvestigationStatus.IN_PROGRESS: 9,
        TestInvestigationStatus.INVESTIGATED: 8,
        TestInvestigationStatus.IGNORED: 7,
    }

    def __init__(self, status: TestInvestigationStatus):
        self._status = status

    def _get_prio(self):
        return self.PRIORITY_MAP.get(self._status, 0)

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestInvestigationStatus):
            return False
        return self._get_prio() == __o._get_prio()

    def __ne__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestInvestigationStatus):
            return False
        return not self.__eq__(__o)

    def __lt__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestInvestigationStatus):
            return False
        return self._get_prio() < __o._get_prio()

    def __gt__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestInvestigationStatus):
            return False
        return self._get_prio() > __o._get_prio()

    def __ge__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestInvestigationStatus):
            return False
        return self._get_prio() >= __o._get_prio()

    def __le__(self, __o: object) -> bool:
        if not isinstance(__o, ComparableTestInvestigationStatus):
            return False
        return self._get_prio() <= __o._get_prio()


def generate_field_status_map(
        last_runs: list[TestRunStatRow],
        field_name = "status",
        container_class = TestStatus,
        cmp_class = ComparableTestStatus
    ) -> dict[int, tuple[str, TestRunStatRow]]:

    status_map = {}
    for run in last_runs:
        run_number = get_build_number(run["build_job_url"])
        match status := status_map.get(run_number):
            case str():
                if cmp_class(container_class(status)) < cmp_class(container_class(run[field_name])):
                    status_map[run_number] = run[field_name]
            case _:
                status_map[run_number] = (run[field_name], run)
    return status_map


class ReleaseStats:
    def __init__(self, release: ArgusRelease) -> None:
        self.release = release
        self.groups: list[GroupStats] = []
        self.status_map = {status: 0 for status in TestStatus}
        self.total_tests = 0
        self.last_status = TestStatus.NOT_PLANNED
        self.last_investigation_status = TestInvestigationStatus.NOT_INVESTIGATED
        self.has_bug_report = False
        self.issues: list[ArgusGithubIssue] = []
        self.comments: list[ArgusTestRunComment] = []
        self.test_schedules: list[ArgusScheduleTest] = []
        self.forced_collection = False
        self.rows = []
        self.all_tests = []

    def to_dict(self) -> dict:
        converted_groups = {str(group.group.id): group.to_dict() for group in self.groups}
        aggregated_investigation_status = {}
        for group in converted_groups.values():
            for investigation_status in TestInvestigationStatus:
                current_status = aggregated_investigation_status.get(investigation_status.value, {})
                result = {
                    status.value: current_status.get(status.value, 0) + group.get(investigation_status.value, {}).get(status, 0)
                    for status in TestStatus
                }
                aggregated_investigation_status[investigation_status.value] = result

        return {
            "release": dict(self.release.items()),
            "groups": converted_groups,
            "total": self.total_tests,
            **self.status_map,
            "disabled": not self.release.enabled,
            "perpetual": self.release.perpetual,
            "lastStatus": self.last_investigation_status,
            "lastInvestigationStatus": self.last_investigation_status,
            "hasBugReport": self.has_bug_report,
            **aggregated_investigation_status
        }

    def collect(self, rows: list[TestRunStatRow], limited=False, force=False) -> None:
        self.forced_collection = force
        if not self.release.enabled and not force:
            return

        if not self.release.perpetual and not limited:
            self.test_schedules = list(ArgusScheduleTest.filter(
                release_id=self.release.id
            ).all())

        self.rows = rows
        if not limited or force:
            self.issues = ArgusGithubIssue.filter(release_id=self.release.id).all()
            self.comments = ArgusTestRunComment.filter(release_id=self.release.id).all()
        self.all_tests = ArgusTest.filter(release_id=self.release.id).all()
        groups: list[ArgusGroup] = ArgusGroup.filter(release_id=self.release.id).all()
        for group in groups:
            if group.enabled:
                stats = GroupStats(group=group, parent_release=self)
                stats.collect(limited=limited)
                self.groups.append(stats)

    def increment_status(self, status=TestStatus.NOT_PLANNED):
        self.total_tests += 1
        self.status_map[TestStatus(status)] += 1
        self.last_status = TestStatus(status)


class GroupStats:
    def __init__(self, group: ArgusGroup, parent_release: ReleaseStats) -> None:
        self.group = group
        self.parent_release = parent_release
        self.status_map = {status: 0 for status in TestStatus}
        self.total_tests = 0
        self.last_status = TestStatus.NOT_PLANNED
        self.last_investigation_status = TestInvestigationStatus.NOT_INVESTIGATED
        self.disabled = False
        self.tests: list[TestStats] = []

    def to_dict(self) -> dict:
        converted_tests = {str(test.test.id): test.to_dict() for test in self.tests}
        investigation_progress = {}
        for test in converted_tests.values():
            progress_for_status = investigation_progress.get(test["investigation_status"], {})
            status_count = progress_for_status.get(test["status"], 0)
            status_count += 1
            progress_for_status[test["status"]] = status_count
            investigation_progress[test["investigation_status"]] = progress_for_status

        return {
            "group": dict(self.group.items()),
            "total": self.total_tests,
            **self.status_map,
            "lastStatus": self.last_status,
            "lastInvestigationStatus": self.last_investigation_status,
            "disabled": self.disabled,
            "tests": converted_tests,
            **investigation_progress
        }

    def collect(self, limited=False):
        tests = [test for test in self.parent_release.all_tests if test.group_id == self.group.id]

        for test in tests:
            if test.enabled:
                stats = TestStats(
                    test=test,
                    parent_group=self,
                    schedules=tuple(
                        schedule for schedule in self.parent_release.test_schedules if schedule.test_id == test.id)
                )
                stats.collect(limited=limited)
                self.tests.append(stats)

    def increment_status(self, status=TestStatus.NOT_PLANNED):
        self.status_map[TestStatus(status)] += 1
        self.total_tests += 1
        self.last_status = TestStatus(status)
        self.parent_release.increment_status(status)


class TestStats:
    def __init__(
        self,
        test: ArgusTest,
        parent_group: GroupStats,
        schedules: list[ArgusScheduleTest] | None = None
    ) -> None:
        self.test = test
        self.parent_group = parent_group
        self.start_time = datetime.fromtimestamp(0)
        self.status = TestStatus.NOT_PLANNED
        self.investigation_status = TestInvestigationStatus.NOT_INVESTIGATED
        self.last_runs: list[dict] = []
        self.has_bug_report = False
        self.has_comments = False
        self.schedules = schedules if schedules else tuple()
        self.is_scheduled = len(self.schedules) > 0
        self.tracked_run_number = None

    def to_dict(self) -> dict:
        return {
            "test": dict(self.test.items()),
            "status": self.status,
            "investigation_status": self.investigation_status,
            "last_runs": self.last_runs,
            "start_time": self.start_time,
            "hasBugReport": self.has_bug_report,
            "hasComments": self.has_comments,
            "buildNumber": self.tracked_run_number,
        }

    def collect(self, limited=False):

        # TODO: Parametrize run limit
        # FIXME: This is only a mitigation, build_number overflows on the build system side.
        last_runs = [r for r in self.parent_group.parent_release.rows if r["build_id"] == self.test.build_system_id][:15]
        last_runs: list[TestRunStatRow] = sorted(
            last_runs, reverse=True, key=lambda r: get_build_number(r["build_job_url"]))
        try:
            last_run = last_runs[0]
        except IndexError:
            self.status = TestStatus.NOT_RUN if self.is_scheduled else TestStatus.NOT_PLANNED
            self.parent_group.increment_status(status=self.status)
            return
        status_map = generate_field_status_map(last_runs)

        worst_case = status_map.get(get_build_number(last_run["build_job_url"]))
        self.status = worst_case[0]
        self.investigation_status = worst_case[1]["investigation_status"]
        self.start_time = last_run["start_time"]

        self.parent_group.increment_status(status=self.status)
        if limited and not self.parent_group.parent_release.forced_collection:
            return

        self.last_runs = [
            {
                "id": run["id"],
                "status": run["status"],
                "build_number": get_build_number(run["build_job_url"]),
                "build_job_name": run["build_id"],
                "start_time": run["start_time"],
                "assignee": run["assignee"],
                "issues": [dict(i.items()) for i in self.parent_group.parent_release.issues if i.run_id == run["id"]],
                "comments": [dict(i.items()) for i in self.parent_group.parent_release.comments if i.test_run_id == run["id"]],
            }
            for run in last_runs
        ]
        try:
            target_run = next(run for run in self.last_runs if run["id"] == worst_case[1]["id"])
        except StopIteration:
            target_run = worst_case[1]
            target_run["issues"] = [dict(i.items()) for i in self.parent_group.parent_release.issues if i.run_id == target_run["id"]]
            target_run["comments"] = [dict(i.items()) for i in self.parent_group.parent_release.comments if i.test_run_id == target_run["id"]]
        self.has_bug_report = len(target_run["issues"]) > 0
        self.parent_group.parent_release.has_bug_report = self.has_bug_report or self.parent_group.parent_release.has_bug_report
        self.has_comments = len(target_run["comments"]) > 0
        self.last_runs = self.last_runs[:5]
        self.tracked_run_number = target_run.get("build_number", get_build_number(target_run.get("build_job_url")))


class ReleaseStatsCollector:
    def __init__(self, release_name: str, release_version: str | None = None) -> None:
        self.database = ScyllaCluster.get()
        self.session = self.database.get_session()
        self.release = None
        self.release_stats = None
        self.release_rows = []
        self.release_name = release_name
        self.release_version = release_version

    def collect(self, limited=False, force=False, include_no_version=False) -> dict:
        self.release: ArgusRelease = ArgusRelease.get(name=self.release_name)
        self.release_rows = [row for plugin in all_plugin_models()
                             for row in plugin.get_stats_for_release(release=self.release)]
        if self.release.dormant and not force:
            return {
                "dormant": True
            }

        if self.release_version:
            if include_no_version:
                expr = lambda row: row["scylla_version"] == self.release_version or not row["scylla_version"] 
            elif self.release_version == "!noVersion":
                expr = lambda row: not row["scylla_version"]
            else:
                expr = lambda row: row["scylla_version"] == self.release_version
        else:
            if include_no_version:
                expr = lambda row: row
            else:
                expr = lambda row: row["scylla_version"]
        self.release_rows = list(filter(expr, self.release_rows))

        self.release_stats = ReleaseStats(release=self.release)
        self.release_stats.collect(rows=self.release_rows, limited=limited, force=force)
        return self.release_stats.to_dict()



import logging
from abc import ABC, abstractmethod
import jenkins
import click
from flask import current_app
from flask.cli import with_appcontext

from argus.backend.db import ScyllaCluster
from argus.backend.models.web import ArgusRelease, ArgusGroup, ArgusTest
from argus.backend.service.release_manager import ReleaseManagerService

LOGGER = logging.getLogger(__name__)


class ArgusTestsMonitor(ABC):
    def __init__(self) -> None:
        self._cluster = ScyllaCluster.get()
        self._existing_releases = list(ArgusRelease.all())
        self._existing_groups = list(ArgusGroup.all())
        self._existing_tests = list(ArgusTest.all())
        self._filtered_groups: list[str] = current_app.config["BUILD_SYSTEM_FILTERED_PREFIXES"]

    def create_release(self, release_name):
        # pylint: disable=no-self-use
        release = ArgusRelease()
        release.name = release_name
        release.save()

        return release

    def create_group(self, release: ArgusRelease, group_name: str, build_id: str, group_pretty_name: str | None = None):
        # pylint: disable=no-self-use
        group = ArgusGroup()
        group.release_id = release.id
        group.name = group_name
        group.build_system_id = build_id
        if group_pretty_name:
            group.pretty_name = group_pretty_name
        group.save()

        return group

    def create_test(self, release: ArgusRelease, group: ArgusGroup,
                    test_name: str, build_id: str, build_url: str) -> ArgusTest:
        # pylint: disable=no-self-use
        test = ArgusTest()
        test.name = test_name
        test.group_id = group.id
        test.release_id = release.id
        test.build_system_id = build_id
        test.build_system_url = build_url
        test.validate_build_system_id()
        test.save()
        ReleaseManagerService().move_test_runs(test)

        return test

    @abstractmethod
    def collect(self):
        raise NotImplementedError()

    def check_filter(self, group_name: str) -> bool:
        for prefix in self._filtered_groups:
            if prefix.lower() in group_name.lower():
                return False

        return True


class JenkinsMonitor(ArgusTestsMonitor):
    def __init__(self) -> None:
        super().__init__()
        self._jenkins = jenkins.Jenkins(url=current_app.config["JENKINS_URL"],
                                        username=current_app.config["JENKINS_USER"],
                                        password=current_app.config["JENKINS_API_TOKEN"])
        self._monitored_releases = current_app.config["JENKINS_MONITORED_RELEASES"]

    def collect(self):
        click.echo("Collecting new tests from jenkins")
        all_jobs = self._jenkins.get_all_jobs()
        all_monitored_folders = [job for job in all_jobs if job["fullname"] in self._monitored_releases]
        for release in all_monitored_folders:
            LOGGER.info("Processing release %s", release["name"])
            try:
                saved_release = ArgusRelease.get(name=release["name"])
                LOGGER.info("Release %s exists", release["name"])
            except ArgusRelease.DoesNotExist:
                LOGGER.warning("Release %s does not exist, creating...", release["name"])
                saved_release = self.create_release(release["name"])
                self._existing_releases.append(saved_release)

            try:
                groups = self.collect_groups_for_release(release["jobs"])
            except KeyError:
                LOGGER.error("Empty release!\n %s", release)
                continue
            folder_stack = [dict(parent_name="", parent_display_name="", group=g) for g in reversed(groups)]
            root_folder = {
                "parent_name": "",
                "parent_display_name": "",
                "group":  {
                    "name": f"{release['fullname']}-root",
                    "displayName": "-- root directory --",
                    "fullname": release["fullname"],
                    "jobs": self.collect_root_folder_jobs(release["jobs"]),
                }
            }
            folder_stack.append(root_folder)
            while len(folder_stack) != 0:
                group_dict = folder_stack.pop()
                group = group_dict["group"]
                LOGGER.info("Processing group %s for release %s", group["name"], saved_release.name)
                try:
                    group_name = group["name"] if not group_dict["parent_name"] else f"{group_dict['parent_name']}-{group['name']}"
                    saved_group = filter(lambda g: g.build_system_id == group["fullname"], self._existing_groups)
                    saved_group = next(saved_group)
                    LOGGER.info("Group %s already exists. (id: %s)", saved_group.build_system_id, saved_group.id)
                except StopIteration:
                    LOGGER.warning(
                        "Group %s for release %s doesn't exist, creating...", group_name, saved_release.name)
                    try:
                        display_name = group.get("displayName", self._jenkins.get_job_info(name=group["fullname"])["displayName"])
                        display_name = display_name if not group_dict[
                            "parent_display_name"] else f"{group_dict['parent_display_name']} - {display_name}"
                    except Exception:
                        display_name = None

                    saved_group = self.create_group(saved_release, group_name, group["fullname"], display_name)
                    self._existing_groups.append(saved_group)

                for job in group["jobs"]:
                    LOGGER.info("Processing job %s for release %s and group %s",
                                job["fullname"], saved_group.name, saved_release.name)
                    saved_test = None
                    if "Folder" in job["_class"]:
                        folder_stack.append(dict(parent_name=saved_group.name,
                                            parent_display_name=saved_group.pretty_name, group=job))
                    if "WorkflowJob" in job["_class"]:
                        try:
                            saved_test = filter(lambda t: t.build_system_id == job["fullname"], self._existing_tests)
                            saved_test = next(saved_test)
                            LOGGER.info("Test %s already exists. (id: %s)", saved_test.build_system_id, saved_test.id)
                        except StopIteration:
                            LOGGER.warning("Test %s for release %s (group %s) doesn't exist, creating...",
                                           job["name"], saved_release.name, saved_group.name)
                            saved_test = self.create_test(
                                saved_release, saved_group, job["name"], job["fullname"], job["url"])
                            self._existing_tests.append(saved_test)

    def collect_groups_for_release(self, jobs):
        # pylint: disable=no-self-use
        groups = [folder for folder in jobs if "Folder" in folder["_class"]]
        groups = [group for group in groups if self.check_filter(group["name"])]

        return groups

    def collect_root_folder_jobs(self, jobs):
        return [job for job in jobs if "WorkflowJob" in job["_class"]]



from datetime import datetime
import json
from argus.backend.models.web import ArgusEvent, ArgusEventTypes


class EventService:
    @staticmethod
    def create_run_event(kind: ArgusEventTypes, body: dict, user_id=None, run_id=None, release_id=None, group_id=None, test_id=None):
        event = ArgusEvent()
        event.release_id = release_id
        event.group_id = group_id
        event.test_id = test_id
        event.user_id = user_id
        event.run_id = run_id
        event.body = json.dumps(body, ensure_ascii=True, separators=(',', ':'))
        event.kind = kind.value
        event.created_at = datetime.utcnow()
        event.save()



from dataclasses import dataclass
from datetime import datetime
from functools import reduce
import re
from uuid import UUID
from cassandra.cqlengine import columns
from argus.backend.db import ScyllaCluster
from argus.backend.models.web import ArgusRelease
from argus.backend.plugins.core import PluginModelBase
from argus.backend.plugins.driver_matrix_tests.udt import TestCollection, TestSuite, TestCase, EnvironmentInfo
from argus.backend.plugins.driver_matrix_tests.raw_types import RawMatrixTestResult
from argus.backend.util.enums import TestStatus


class DriverMatrixPluginError(Exception):
    pass


@dataclass(init=True, repr=True, frozen=True)
class DriverMatrixRunSubmissionRequest():
    schema_version: str
    run_id: str
    job_name: str
    job_url: str
    test_environment: dict[str, str]
    matrix_results: list[RawMatrixTestResult]


class DriverTestRun(PluginModelBase):
    _plugin_name = "driver-matrix-tests"
    __table_name__ = "driver_test_run"
    scylla_version = columns.Text()
    test_collection = columns.List(value_type=columns.UserDefinedType(user_type=TestCollection))
    environment_info = columns.List(value_type=columns.UserDefinedType(user_type=EnvironmentInfo))


    _artifact_fnames = {
        "cpp": r"TEST-(?P<driver_name>[\w]*)-(?P<version>[\d\.-]*)",
        "gocql": r"xunit\.(?P<driver_name>[\w]*)\.(?P<proto>v\d)\.(?P<version>[v\d\.]*)",
        "python": r"pytest\.(?P<driver_name>[\w]*)\.(?P<proto>v\d)\.(?P<version>[\d\.]*)",
        "java": r"TEST-(?P<version>[\d\.\w-]*)",
    }

    @classmethod
    def _stats_query(cls) -> str:
        return ("SELECT id, test_id, group_id, release_id, status, start_time, build_job_url, build_id, "
                f"assignee, end_time, investigation_status, heartbeat, scylla_version FROM {cls.table_name()} WHERE release_id = ?")

    @classmethod
    def get_distinct_product_versions(cls, release: ArgusRelease) -> list[str]:
        cluster = ScyllaCluster.get()
        statement = cluster.prepare(f"SELECT scylla_version FROM {cls.table_name()} WHERE release_id = ?")
        rows = cluster.session.execute(query=statement, parameters=(release.id,))
        unique_versions = {r["scylla_version"] for r in rows if r["scylla_version"]}

        return sorted(list(unique_versions), reverse=True)

    @classmethod
    def load_test_run(cls, run_id: UUID) -> 'DriverTestRun':
        return cls.get(id=run_id)

    @classmethod
    def parse_driver_name(cls, raw_file_name: str) -> str:
        for test, pattern in cls._artifact_fnames.items():
            match = re.match(pattern, raw_file_name)
            if not match:
                continue
            driver_info = match.groupdict()
            if test == "java":
                version = driver_info["version"]
                return "scylla" if len(version.split(".")) > 3 or "scylla" in version else "datastax"
            else:
                return driver_info["driver_name"]
        return "unknown_driver"


    @classmethod
    def submit_run(cls, request_data: dict) -> 'DriverTestRun':
        req = DriverMatrixRunSubmissionRequest(**request_data)
        run = cls()
        run.id = req.run_id  # pylint: disable=invalid-name
        run.build_id = req.job_name
        run.build_job_url = req.job_url
        run.assign_categories()
        try:
            run.assignee = run.get_scheduled_assignee()
        except Exception:  # pylint: disable=broad-except
            run.assignee = None
        for key, value in req.test_environment.items():
            env_info = EnvironmentInfo()
            env_info.key = key
            env_info.value = value
            run.environment_info.append(env_info)

        run.scylla_version = req.test_environment.get("scylla-version")
        run.test_collection = []

        for result in req.matrix_results:
            collection = TestCollection()
            collection.name = result.get("name")
            collection.driver = cls.parse_driver_name(collection.name)
            collection.tests_total = result.get("tests")
            collection.failures = result.get("failures")
            collection.errors = result.get("errors")
            collection.skipped = result.get("skipped")
            collection.passed = result.get("passed")
            collection.disabled = result.get("disabled")
            collection.time = result.get("time")
            timestamp = result.get("timestamp")
            # TODO: Not needed once python>=3.11
            timestamp = timestamp[0:-1] if timestamp[-1] == "Z" else timestamp
            collection.timestamp = datetime.fromisoformat(timestamp)
            for raw_suite in result.get("suites"):
                suite = TestSuite()
                suite.name = raw_suite.get("name")
                suite.tests_total = raw_suite.get("tests")
                suite.failures = raw_suite.get("failures")
                suite.errors = raw_suite.get("errors")
                suite.skipped = raw_suite.get("skipped")
                suite.passed = raw_suite.get("passed")
                suite.disabled = raw_suite.get("disabled")
                suite.time = raw_suite.get("time")

                for raw_case in raw_suite.get("cases"):
                    case = TestCase()
                    case.name = raw_case.get("name")
                    case.status = raw_case.get("status")
                    case.time = raw_case.get("time")
                    case.classname = raw_case.get("classname")
                    case.message = raw_case.get("message")
                    suite.cases.append(case)

                collection.suites.append(suite)
            run.test_collection.append(collection)

        run.status = run._determine_run_status().value
        run.save()
        return run

    def get_resources(self) -> list:
        return []

    def get_nemeses(self) -> list:
        return []

    def _determine_run_status(self):
        if len(self.test_collection) < 2:
            return TestStatus.FAILED

        driver_types = {collection.driver for collection in self.test_collection}
        if len(driver_types) <= 1:
            return TestStatus.FAILED

        failure_count = reduce(lambda acc, val: acc + (val.failures + val.errors), self.test_collection, 0)
        if failure_count > 0:
            return TestStatus.FAILED

        return TestStatus.PASSED

    def change_status(self, new_status: TestStatus):
        raise DriverMatrixPluginError("This method is obsolete. Status is now determined on submission.")

    def get_events(self) -> list:
        return []

    def submit_product_version(self, version: str):
        self.scylla_version = version

    def finish_run(self, payload: dict = None):
        self.end_time = datetime.utcnow()

    def submit_logs(self, logs: list[dict]):
        pass



from datetime import datetime
from uuid import UUID
from cassandra.cqlengine import columns
from cassandra.cqlengine.models import Model
from argus.backend.db import ScyllaCluster
from argus.backend.models.web import ArgusRelease
from argus.backend.plugins.core import PluginModelBase
from argus.backend.plugins.generic.types import GenericRunFinishRequest, GenericRunSubmitRequest
from argus.backend.util.enums import TestStatus


class GenericPluginException(Exception):
    pass


class GenericRun(PluginModelBase):
    _plugin_name = "generic"
    __table_name__ = "generic_run"
    logs = columns.Map(key_type=columns.Text(), value_type=columns.Text())
    started_by = columns.Text()
    # TODO: Legacy field name, should be renamed to product_version and abstracted
    scylla_version = columns.Text()

    @classmethod
    def _stats_query(cls) -> str:
        return ("SELECT id, test_id, group_id, release_id, status, start_time, build_job_url, build_id, "
                f"assignee, end_time, investigation_status, heartbeat, scylla_version FROM {cls.table_name()} WHERE release_id = ?")

    @classmethod
    def get_distinct_product_versions(cls, release: ArgusRelease, cluster: ScyllaCluster = None) -> list[str]:
        if not cluster:
            cluster = ScyllaCluster.get()
        statement = cluster.prepare(f"SELECT scylla_version FROM {cls.table_name()} WHERE release_id = ?")
        rows = cluster.session.execute(query=statement, parameters=(release.id,))
        unique_versions = {r["scylla_version"] for r in rows if r["scylla_version"]}

        return sorted(list(unique_versions), reverse=True)

    def submit_product_version(self, version: str):
        self.scylla_version = version
        self.set_product_version(version)

    @classmethod
    def load_test_run(cls, run_id: UUID) -> 'GenericRun':
        return cls.get(id=run_id)

    @classmethod
    def submit_run(cls, request_data: GenericRunSubmitRequest) -> 'GenericRun':
        try:
            run = cls.get(id=request_data["run_id"])
            raise GenericPluginException(f"Run with UUID {request_data['run_id']} already exists.", request_data["run_id"])
        except cls.DoesNotExist:
            pass
        run = cls()
        run.start_time = datetime.utcnow()
        run.build_id = request_data["build_id"]
        run.started_by = request_data["started_by"]
        run.id = request_data["run_id"]
        run.build_job_url = request_data["build_url"]
        run.assign_categories()
        try:
            run.assignee = run.get_scheduled_assignee()
        except Model.DoesNotExist:
            run.assignee = None
        if version := request_data.get("scylla_version"):
            run.submit_product_version(version)
        run.status = TestStatus.RUNNING.value
        run.save()

        return run

    def finish_run(self, payload: GenericRunFinishRequest = None):
        self.end_time = datetime.utcnow()
        self.status = TestStatus(payload["status"]).value
        if version := payload.get("scylla_version"):
            self.submit_product_version(version)



from typing import TypedDict

class GenericRunSubmitRequest(TypedDict):
    build_id: str
    build_url: str
    run_id: str
    started_by: str
    scylla_version: str | None


class GenericRunFinishRequest(TypedDict):
    status: str
    scylla_version: str | None



import logging
from argus.client.base import ArgusClient


LOGGER = logging.getLogger(__name__)

class ArgusGenericClient(ArgusClient):
    test_type = "generic"
    schema_version: None = "v1"
    def __init__(self, auth_token: str, base_url: str, api_version="v1") -> None:
        super().__init__(auth_token, base_url, api_version)

    def submit_generic_run(self, build_id: str, run_id: str, started_by: str, build_url: str, scylla_version: str | None = None):
        request_body = {
            "build_id": build_id,
            "run_id": run_id,
            "started_by": started_by,
            "build_url": build_url,
            "scylla_version": scylla_version
        }

        response = self.submit_run(run_type=self.test_type, run_body=request_body)
        self.check_response(response)


    def finalize_generic_run(self, run_id: str, status: str, scylla_version: str | None = None):
        response = self.finalize_run(run_type=self.test_type, run_id=run_id, body={
            "status": status,
            "scylla_version": scylla_version,
        })
        self.check_response(response)


import logging
from uuid import UUID
from flask import (
    Blueprint,
    request
)

from argus.backend.error_handlers import handle_api_exception
from argus.backend.service.testrun import TestRunService
from argus.backend.service.user import api_login_required
from argus.backend.util.common import get_payload
from argus.backend.util.enums import TestInvestigationStatus, TestStatus

bp = Blueprint('testrun_api', __name__, 'testrun')
bp.register_error_handler(Exception, handle_api_exception)
LOGGER = logging.getLogger(__name__)


@bp.route("/test/<string:test_id>/runs")
@api_login_required
def get_runs_for_test(test_id: str):
    limit = request.args.get("limit")
    if not limit:
        limit = 10
    else:
        limit = int(limit)
    additional_runs = [UUID(run) for run in request.args.getlist('additionalRuns[]')]
    service = TestRunService()

    runs = service.get_runs_by_test_id(test_id=UUID(test_id), additional_runs=additional_runs, limit=limit)

    return {
        "status": "ok",
        "response": runs
    }


@bp.route("/run/<string:run_type>/<string:run_id>")
@api_login_required
def get_testrun(run_type: str, run_id: str):
    run_id = UUID(run_id)
    service = TestRunService()
    test_run = service.get_run(run_type=run_type, run_id=run_id)

    return {
        "status": "ok",
        "response": test_run
    }


@bp.route("/run/<string:run_id>/activity")
@api_login_required
def test_run_activity(run_id: str):
    run_id = UUID(run_id)
    service = TestRunService()
    activity = service.get_run_events(run_id=run_id)

    return {
        "status": "ok",
        "response": activity
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/status/set", methods=["POST"])
@api_login_required
def set_testrun_status(test_id: str, run_id: str):
    payload = get_payload(client_request=request)
    new_status = payload.get("status")
    if not new_status:
        raise Exception("Status not specified in the request")
    service = TestRunService()
    result = service.change_run_status(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        new_status=TestStatus(new_status)
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/investigation_status/set", methods=["POST"])
@api_login_required
def set_testrun_investigation_status(test_id: str, run_id: str):
    payload = get_payload(client_request=request)
    new_status = payload.get("investigation_status")
    if not new_status:
        raise Exception("Status not specified in the request")
    service = TestRunService()
    result = service.change_run_investigation_status(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        new_status=TestInvestigationStatus(new_status),
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/assignee/set", methods=["POST"])
@api_login_required
def set_testrun_assignee(test_id: str, run_id: str):
    payload = get_payload(client_request=request)
    assignee = payload.get("assignee")
    if not assignee:
        raise Exception("Assignee not specified in the request")
    service = TestRunService()
    result = service.change_run_assignee(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        new_assignee=UUID(assignee) if assignee != TestRunService.ASSIGNEE_PLACEHOLDER else None,
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/issues/submit", methods=["POST"])
@api_login_required
def issues_submit(test_id: str, run_id: str):
    payload = get_payload(client_request=request)
    service = TestRunService()
    submit_result = service.submit_github_issue(
        issue_url=payload["issue_url"],
        test_id=UUID(test_id),
        run_id=UUID(run_id)
    )

    return {
        "status": "ok",
        "response": submit_result
    }


@bp.route("/issues/get", methods=["GET"])
@api_login_required
def issues_get():
    filter_key = request.args.get("filterKey")
    if not filter_key:
        raise Exception("Filter key not provided in the request")
    key_value = request.args.get("id")
    if not key_value:
        raise Exception("Id wasn't provided in the request")
    key_value = UUID(key_value)
    aggregate_by_issue = request.args.get("aggregateByIssue")
    aggregate_by_issue = bool(int(aggregate_by_issue)) if aggregate_by_issue else False
    service = TestRunService()
    issues = service.get_github_issues(
        filter_key=filter_key,
        filter_id=key_value,
        aggregate_by_issue=aggregate_by_issue
    )

    return {
        "status": "ok",
        "response": issues
    }


@bp.route("/issues/delete", methods=["POST"])
@api_login_required
def issues_delete():
    payload = get_payload(client_request=request)
    service = TestRunService()
    result = service.delete_github_issue(issue_id=payload["issue_id"])

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/run/<string:run_id>/comments", methods=["GET"])
@api_login_required
def get_testrun_comments(run_id: str):
    service = TestRunService()
    comments = service.get_run_comments(
        run_id=UUID(run_id)
    )

    return {
        "status": "ok",
        "response": comments
    }


@bp.route("/comment/<string:comment_id>/get", methods=["GET"])
@api_login_required
def get_single_comment(comment_id: str):
    service = TestRunService()
    comment = service.get_run_comment(
        comment_id=UUID(comment_id)
    )

    return {
        "status": "ok",
        "response": comment
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/comments/submit", methods=["POST"])
@api_login_required
def submit_testrun_comment(test_id: str, run_id: str):
    payload = get_payload(request)
    service = TestRunService()
    result = service.post_run_comment(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        message=payload["message"],
        reactions=payload["reactions"],
        mentions=payload["mentions"],
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/comment/<string:comment_id>/update", methods=["POST"])
@api_login_required
def test_run_update_comment(test_id: str, run_id: str, comment_id: str):
    payload = get_payload(request)
    service = TestRunService()
    result = service.update_run_comment(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        comment_id=UUID(comment_id),
        message=payload["message"],
        reactions=payload["reactions"],
        mentions=payload["mentions"],
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/comment/<string:comment_id>/delete", methods=["POST"])
@api_login_required
def test_run_delete_comment(test_id: str, run_id: str, comment_id: str):
    service = TestRunService()
    result = service.delete_run_comment(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        comment_id=UUID(comment_id),
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/terminate_stuck_runs", methods=["POST"])
@api_login_required
def sct_terminate_stuck_runs():
    result = TestRunService().terminate_stuck_runs()
    return {
        "status": "ok",
        "response": {
            "total": result
        }
    }


@bp.route("/ignore_jobs", methods=["POST"])
@api_login_required
def ignore_jobs():
    payload = get_payload(request)
    service = TestRunService()

    result = service.ignore_jobs(test_id=payload["testId"], reason=payload["reason"])

    return {
        "status": "ok",
        "response": {
            "affectedJobs": result
        }
    }



from flask import Blueprint, request

from argus.backend.error_handlers import handle_api_exception
from argus.backend.service.user import api_login_required
from argus.backend.plugins.sct.service import SCTService
from argus.backend.util.common import get_payload

bp = Blueprint("sct_api", __name__, url_prefix="/sct")
bp.register_error_handler(Exception, handle_api_exception)


@bp.route("/<string:run_id>/packages/submit", methods=["POST"])
@api_login_required
def sct_submit_packages(run_id: str):
    payload = get_payload(request)
    result = SCTService.submit_packages(run_id=run_id, packages=payload["packages"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:run_id>/screenshots/submit", methods=["POST"])
@api_login_required
def sct_submit_screenshots(run_id: str):
    payload = get_payload(request)
    result = SCTService.submit_screenshots(run_id=run_id, screenshot_links=payload["screenshot_links"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:run_id>/sct_runner/set", methods=["POST"])
@api_login_required
def sct_set_runner(run_id: str):
    payload = get_payload(request)
    result = SCTService.set_sct_runner(
        run_id=run_id,
        public_ip=payload["public_ip"],
        private_ip=payload["private_ip"],
        region=payload["region"],
        backend=payload["backend"]
    )
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:run_id>/resource/create", methods=["POST"])
@api_login_required
def sct_resource_create(run_id: str):
    payload = get_payload(request)
    result = SCTService.create_resource(run_id=run_id, resource_details=payload["resource"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:run_id>/resource/<string:resource_name>/terminate", methods=["POST"])
@api_login_required
def sct_resource_terminate(run_id: str, resource_name: str):
    payload = get_payload(request)
    result = SCTService.terminate_resource(run_id=run_id, resource_name=resource_name, reason=payload["reason"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:run_id>/resource/<string:resource_name>/shards", methods=["POST"])
@api_login_required
def sct_resource_update_shards(run_id: str, resource_name: str):
    payload = get_payload(request)
    result = SCTService.update_resource_shards(run_id=run_id, resource_name=resource_name, new_shards=payload["shards"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:run_id>/nemesis/submit", methods=["POST"])
@api_login_required
def sct_nemesis_submit(run_id: str):
    payload = get_payload(request)
    result = SCTService.submit_nemesis(run_id=run_id, nemesis_details=payload["nemesis"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:run_id>/nemesis/finalize", methods=["POST"])
@api_login_required
def sct_nemesis_finalize(run_id: str):
    payload = get_payload(request)
    result = SCTService.finalize_nemesis(run_id=run_id, nemesis_details=payload["nemesis"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:run_id>/events/submit", methods=["POST"])
@api_login_required
def sct_events_submit(run_id: str):
    payload = get_payload(request)
    result = SCTService.submit_events(run_id=run_id, events=payload["events"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/<string:run_id>/gemini/submit", methods=["POST"])
@api_login_required
def sct_gemini_results_submit(run_id: str):
    payload = get_payload(request)
    result = SCTService.submit_gemini_results(run_id=run_id, gemini_data=payload["gemini_data"])
    return {
        "status": "ok",
        "response": result
    }

@bp.route("/<string:run_id>/performance/submit", methods=["POST"])
@api_login_required
def sct_performance_results_submit(run_id: str):
    payload = get_payload(request)
    result = SCTService.submit_performance_results(run_id=run_id, performance_results=payload["performance_results"])
    return {
        "status": "ok",
        "response": result
    }

@bp.route("/<string:run_id>/performance/history", methods=["GET"])
@api_login_required
def sct_get_performance_history(run_id: str):
    result = SCTService.get_performance_history_for_test(run_id=run_id)
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/release/<string:release_name>/kernels", methods=["GET"])
@api_login_required
def sct_get_kernel_report(release_name: str):
    result = SCTService.get_scylla_version_kernels_report(release_name=release_name)
    return {
        "status": "ok",
        "response": result
    }



from datetime import datetime
from uuid import UUID, uuid4
from cassandra.cqlengine import columns
from cassandra.cqlengine.usertype import UserType
from cassandra.cqlengine.models import Model
from argus.backend.db import ScyllaCluster
from argus.backend.models.web import ArgusRelease
from argus.backend.plugins.core import PluginModelBase
from argus.backend.plugins.sirenada.types import RawSirenadaRequest, SirenadaPluginException
from argus.backend.util.enums import TestStatus


class SirenadaTest(UserType):
    test_name = columns.Text()
    class_name = columns.Text()
    file_name = columns.Text()
    browser_type = columns.Text()
    cluster_type = columns.Text()
    status = columns.Text()
    duration = columns.Float()
    message = columns.Text()
    start_time = columns.DateTime()
    stack_trace = columns.Text()
    screenshot_file = columns.Text()
    s3_folder_id = columns.Text()
    requests_file = columns.Text()
    sirenada_test_id = columns.Text()
    sirenada_user = columns.Text()
    sirenada_password = columns.Text()



class SirenadaRun(PluginModelBase):
    _plugin_name = "sirenada"
    __table_name__ = "sirenada_run"
    logs = columns.Map(key_type=columns.Text(), value_type=columns.Text())
    # TODO: Legacy field name, should be renamed to product_version and abstracted
    scylla_version = columns.Text()
    region = columns.Text()
    sirenada_test_ids = columns.List(value_type=columns.Text())
    s3_folder_ids = columns.List(value_type=columns.Tuple(columns.Text(), columns.Text()))
    browsers = columns.List(value_type=columns.Text())
    clusters = columns.List(value_type=columns.Text())
    sct_test_id = columns.UUID()
    results = columns.List(value_type=columns.UserDefinedType(user_type=SirenadaTest))

    @classmethod
    def _stats_query(cls) -> str:
        return ("SELECT id, test_id, group_id, release_id, status, start_time, build_job_url, build_id, "
                f"assignee, end_time, investigation_status, heartbeat, scylla_version FROM {cls.table_name()} WHERE release_id = ?")

    @classmethod
    def get_distinct_product_versions(cls, release: ArgusRelease, cluster: ScyllaCluster = None) -> list[str]:
        if not cluster:
            cluster = ScyllaCluster.get()
        statement = cluster.prepare(f"SELECT scylla_version FROM {cls.table_name()} WHERE release_id = ?")
        rows = cluster.session.execute(query=statement, parameters=(release.id,))
        unique_versions = {r["scylla_version"] for r in rows if r["scylla_version"]}

        return sorted(list(unique_versions), reverse=True)

    def submit_product_version(self, version: str):
        self.scylla_version = version

    def submit_logs(self, logs: dict[str, str]):
        raise SirenadaPluginException("Log submission is not supported for Sirenada")

    def finish_run(self, payload: dict = None):
        raise SirenadaPluginException("Sirenada runs do not need finalization")

    @classmethod
    def load_test_run(cls, run_id: UUID) -> 'SirenadaRun':
        return cls.get(id=run_id)

    @classmethod
    def submit_run(cls, request_data: RawSirenadaRequest) -> 'SirenadaRun':
        try:
            run = cls.get(id=UUID(request_data["run_id"]))
        except cls.DoesNotExist:
            run = cls()
            run.id = request_data["run_id"] # FIXME: Validate pls
            run.build_id = request_data["build_id"]
            run.start_time = datetime.utcnow()
            run.assign_categories()
            run.build_job_url = request_data["build_job_url"]
            run.region = request_data["region"]
            run.status = TestStatus.PASSED.value
            try:
                run.assignee = run.get_scheduled_assignee()
            except Model.DoesNotExist:
                run.assignee = None

        for raw_case in request_data["results"]:
            case = SirenadaTest(**raw_case)
            if case.status in ["failed", "error"] and run.status not in [TestStatus.FAILED.value, TestStatus.ABORTED.value]:
                run.status = TestStatus.FAILED.value
            run.results.append(case)

            if case.sirenada_test_id not in run.sirenada_test_ids:
                run.sirenada_test_ids.append(case.sirenada_test_id)

            if case.browser_type not in run.browsers:
                run.browsers.append(case.browser_type)

            if case.cluster_type not in run.clusters:
                run.clusters.append(case.cluster_type)

            if (case.s3_folder_id, case.sirenada_test_id) not in run.s3_folder_ids and case.s3_folder_id:
                run.s3_folder_ids.append((case.s3_folder_id, case.sirenada_test_id))

        run.save()

        return run



import logging
from uuid import UUID
from flask import (
    Blueprint,
    request,
    Request,
)
from argus.backend.error_handlers import handle_api_exception
from argus.backend.service.release_manager import ReleaseEditPayload, ReleaseManagerService
from argus.backend.service.user import api_login_required, check_roles
from argus.backend.models.web import UserRoles

bp = Blueprint('admin_api', __name__, url_prefix='/api/v1')
LOGGER = logging.getLogger(__name__)
bp.register_error_handler(Exception, handle_api_exception)


def get_payload(client_request: Request):
    if not client_request.is_json:
        raise Exception(
            "Content-Type mismatch, expected application/json, got:", client_request.content_type)
    request_payload = client_request.get_json()

    return request_payload


@bp.route("/", methods=["GET"])
@check_roles(UserRoles.Admin)
@api_login_required
def index():
    return {
        "version": "v1"
    }


@bp.route("/release/create", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def create_release():
    payload = get_payload(request)
    release = ReleaseManagerService().create_release(**payload)

    return {
        "status": "ok",
        "response": {
            "new_release": release
        }
    }


@bp.route("/release/set_perpetual", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def set_release_perpetual():
    payload = get_payload(request)
    result = ReleaseManagerService().set_release_perpetuality(**payload)
    return {
        "status": "ok",
        "response": {
            "updated": result
        }
    }


@bp.route("/release/set_state", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def set_release_state():
    payload = get_payload(request)
    result = ReleaseManagerService().set_release_state(**payload)

    return {
        "status": "ok",
        "response": {
            "updated": result
        }
    }


@bp.route("/release/set_dormant", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def set_release_dormancy():
    payload = get_payload(request)
    result = ReleaseManagerService().set_release_dormancy(**payload)

    return {
        "status": "ok",
        "response": {
            "updated": result
        }
    }


@bp.route("/release/edit", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def edit_release():
    payload: ReleaseEditPayload = get_payload(request)
    result = ReleaseManagerService().edit_release(payload)

    return {
        "status": "ok",
        "response": {
            "updated": result
        }
    }


@bp.route("/release/delete", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def delete_release():
    payload = get_payload(request)
    result = ReleaseManagerService().delete_release(release_id=payload["releaseId"])

    return {
        "status": "ok",
        "response": {
            "deleted": result
        }
    }


@bp.route("/group/create", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def create_group():
    payload = get_payload(request)
    group = ReleaseManagerService().create_group(**payload)

    return {
        "status": "ok",
        "response": {
            "new_group": group
        }
    }


@bp.route("/group/update", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def update_group():
    payload = get_payload(request)
    result = ReleaseManagerService().update_group(**payload)
    return {
        "status": "ok",
        "response": {
            "updated": result
        }
    }


@bp.route("/group/delete", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def delete_group():
    payload = get_payload(request)
    result = ReleaseManagerService().delete_group(**payload)

    return {
        "status": "ok",
        "response": {
            "deleted": result
        }
    }


@bp.route("/test/create", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def create_test():
    payload = get_payload(request)
    test = ReleaseManagerService().create_test(**payload)
    return {
        "status": "ok",
        "response": {
            "new_test": test
        }
    }


@bp.route("/test/update", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def update_test():
    payload = get_payload(request)
    result = ReleaseManagerService().update_test(**payload)
    return {
        "status": "ok",
        "response": {
            "updated": result
        }
    }


@bp.route("/test/batch_move", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def batch_move_tests():
    payload = get_payload(request)
    result = ReleaseManagerService().batch_move_tests(**payload)
    return {
        "status": "ok",
        "response": {
            "moved": result
        }
    }


@bp.route("/test/delete", methods=["POST"])
@check_roles(UserRoles.Admin)
@api_login_required
def delete_test():
    payload = get_payload(request)
    result = ReleaseManagerService().delete_test(**payload)

    return {
        "status": "ok",
        "response": {
            "deleted": result
        }
    }


@bp.route("/releases/get", methods=["GET"])
@check_roles(UserRoles.Admin)
@api_login_required
def get_releases():
    releases = ReleaseManagerService().get_releases()
    return {
        "status": "ok",
        "response": releases
    }


@bp.route("/groups/get", methods=["GET"])
@check_roles(UserRoles.Admin)
@api_login_required
def get_groups_for_release():
    release_id = request.args.get("releaseId")
    if not release_id:
        raise Exception("ReleaseId not provided in the request")
    groups = ReleaseManagerService().get_groups(release_id=UUID(release_id))

    return {
        "status": "ok",
        "response": groups
    }


@bp.route("/tests/get", methods=["GET"])
@check_roles(UserRoles.Admin)
@api_login_required
def get_tests_for_group():
    group_id = request.args.get("groupId")
    if not group_id:
        raise Exception("GroupId not provided in the request")
    tests = ReleaseManagerService().get_tests(group_id=UUID(group_id))
    return {
        "status": "ok",
        "response": tests
    }



from flask import (
    Blueprint,
    render_template,
)
from argus.backend.service.user import login_required

bp = Blueprint('notifications', __name__, url_prefix='/notifications')


@bp.route("/")
@login_required
def index():
    return render_template("profile_notifications.html.j2")



from time import time
from cassandra.cqlengine.usertype import UserType
from cassandra.cqlengine import columns

from argus.backend.util.enums import ResourceState


class PackageVersion(UserType):
    __type_name__ = "PackageVersion_v2"
    name = columns.Text()
    version = columns.Text()
    date = columns.Text()
    revision_id = columns.Text()
    build_id = columns.Text()


class CloudInstanceDetails(UserType):
    __type_name__ = "CloudInstanceDetails_v3"
    provider = columns.Text()
    region = columns.Text()
    public_ip = columns.Text()
    private_ip = columns.Text()
    dc_name = columns.Text()
    rack_name = columns.Text()
    creation_time = columns.Integer(default=lambda: int(time()))
    termination_time = columns.Integer(default=lambda: 0)
    termination_reason = columns.Text(default=lambda: "")
    shards_amount = columns.Integer(default=lambda: 0)


class CloudNodesInfo(UserType):
    __type_name__ = "CloudNodesInfo"
    image_id = columns.Text()
    instance_type = columns.Text()
    node_amount = columns.Integer()
    post_behaviour = columns.Text()


class CloudSetupDetails(UserType):
    __type_name__ = "CloudSetupDetails"
    db_node = columns.UserDefinedType(user_type=CloudNodesInfo)
    loader_node = columns.UserDefinedType(user_type=CloudNodesInfo)
    monitor_node = columns.UserDefinedType(user_type=CloudNodesInfo)
    backend = columns.Text()


class CloudResource(UserType):
    __type_name__ = "CloudResource_v3"
    name = columns.Text()
    state = columns.Text(default=lambda: ResourceState.RUNNING)
    resource_type = columns.Text()
    instance_info = columns.UserDefinedType(user_type=CloudInstanceDetails)

    def get_instance_info(self) -> CloudInstanceDetails:
        return self.instance_info


class EventsBySeverity(UserType):
    __type_name__ = "EventsBySeverity"
    severity = columns.Text()
    event_amount = columns.Integer()
    last_events = columns.List(value_type=columns.Text())


class NodeDescription(UserType):
    __type_name__ = "NodeDescription"
    name = columns.Text()
    ip = columns.Text()
    shards = columns.Integer()


class NemesisRunInfo(UserType):
    __type_name__ = "NemesisRunInfo"

    class_name = columns.Text()
    name = columns.Text()
    duration = columns.Integer()
    target_node = columns.UserDefinedType(user_type=NodeDescription)
    status = columns.Text()
    start_time = columns.Integer()
    end_time = columns.Integer()
    stack_trace = columns.Text()


class PerformanceHDRHistogram(UserType):
    start_time = columns.Integer()
    percentile_90 = columns.Float()
    percentile_50 = columns.Float()
    percentile_99_999 = columns.Float()
    percentile_95 = columns.Float()
    end_time = columns.Float()
    percentile_99_99 = columns.Float()
    percentile_99 = columns.Float()
    stddev = columns.Float()
    percentile_99_9 = columns.Float()



from typing import TypedDict

class RawHDRHistogram(TypedDict):
    start_time: int
    percentile_90: float
    percentile_50: float
    percentile_99_999: float
    percentile_95: float
    end_time: float
    percentile_99_99: float
    percentile_99: float
    stddev: float
    percentile_99_9: float

class GeminiResultsRequest(TypedDict):
    oracle_nodes_count: int
    oracle_node_ami_id: str
    oracle_node_instance_type: str
    oracle_node_scylla_version: str
    gemini_command: str
    gemini_version: str
    gemini_status: str
    gemini_seed: str
    gemini_write_ops: int
    gemini_write_errors: int
    gemini_read_ops: int
    gemini_read_errors: int

class PerformanceResultsRequest(TypedDict):
    test_name: str
    stress_cmd: str
    perf_op_rate_average: float
    perf_op_rate_total: float
    perf_avg_latency_99th: float
    perf_avg_latency_mean: float
    perf_total_errors: str

    histograms: list[dict[str, RawHDRHistogram]] | None



import smtplib
from typing import List, Set
from smtplib import SMTPException

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from flask import current_app
from flask import render_template

class Email:
    #  pylint: disable=too-many-instance-attributes
    """
    Responsible for sending emails
    """
    _attachments_size_limit = 10485760  # 10Mb = 20 * 1024 * 1024
    _body_size_limit = 26214400  # 25Mb = 20 * 1024 * 1024

    def __init__(self, init_connection=True):
        self.sender: str = ""
        self._password: str = ""
        self._user: str = ""
        self._server_host: str = ""
        self._server_port: int = 0
        self._connection: smtplib.SMTP | None = None
        self._retrieve_credentials()
        if init_connection:
            self._connect()

    def _retrieve_credentials(self):
        self.sender = current_app.config["EMAIL_SENDER"]
        self._password = current_app.config["EMAIL_SENDER_PASS"]
        self._user = current_app.config["EMAIL_SENDER_USER"]
        self._server_host = current_app.config["EMAIL_SERVER"]
        self._server_port = int(current_app.config["EMAIL_SERVER_PORT"])

    def _connect(self):
        try:
            self._connection = smtplib.SMTP(host=self._server_host, port=self._server_port)
            self._connection.ehlo()
            self._connection.starttls()
            self._connection.login(user=self._user, password=self._password)
        except SMTPException as details:
            current_app.logger.error("Failed to initialize smtp session %s", details)

    def _is_connection_open(self):
        if not self._connection:
            return False
        try:
            status, _ =  self._connection.noop()
        except SMTPException:
            status = -1

        return True if status == 250 else False

    def _prepare_email(self, subject:str,
                      content: str,
                      recipients: List[str],
                      html: bool = True):  # pylint: disable=too-many-arguments
        msg = MIMEMultipart()
        msg['subject'] = subject
        msg['from'] = self.sender
        assert recipients, "No recipients provided"
        msg['to'] = ','.join(recipients)
        if html:
            text_part = MIMEText(content, "html")
        else:
            text_part = MIMEText(content, "plain")
        msg.attach(text_part)
        email = msg.as_string()
        return email

    def send(self, subject, content, recipients, html=True):  # pylint: disable=too-many-arguments
        """
        :param subject: text
        :param content: text/html
        :param recipients: iterable, list of recipients
        :param html: True/False
        :param files: paths of the files that will be attached to the email
        :return:
        """
        email = self._prepare_email(subject, content, recipients, html)
        self._send_email(recipients, email)

    def _send_email(self, recipients, email):
        if not self._is_connection_open():
            self._connect()
        self._connection.sendmail(self.sender, recipients, email)

    def __del__(self):
        if self._connection:
            self._connection.quit()



import logging
from uuid import UUID

from flask.globals import g
from argus.backend.db import ScyllaCluster
from argus.backend.models.web import Team, User

LOGGER = logging.getLogger(__name__)


class TeamManagerException(Exception):
    pass


class TeamManagerService:
    # pylint: disable=no-self-use
    def __init__(self, database: ScyllaCluster | None = None) -> None:
        if not database:
            database = ScyllaCluster.get()

        self.session = database.get_session()
        self.database = database

    def create_team(self, name: str, leader: UUID, members: list[UUID]):
        team = Team()
        team.name = name
        team.leader = leader
        team.members = [leader, *members]

        team.save()
        return team

    def get_teams_for_user(self, user_id: UUID) -> list[Team]:
        return list(Team.filter(leader=user_id).all())

    def get_team_by_id(self, team_id: UUID) -> Team:
        try:
            return Team.get(id=team_id)
        except Team.DoesNotExist as exc:
            raise TeamManagerException(f"Team {team_id} does not exist", team_id) from exc

    def edit_team(self, team_id: UUID, name: str, members: list[UUID]):
        user: User = g.user
        try:
            team: Team = Team.get(id=team_id)
            if team.leader != user.id:
                raise TeamManagerException(f"Cannot edit team \"{team.name}\" as it doesn't belong to the user.")
            team.name = name
            team.members = [team.leader, *members] if team.leader not in members else members

            team.save()

            return team
        except Team.DoesNotExist as exc:
            raise TeamManagerException(f"Team {team_id} doesn't exist!", team_id) from exc

    def edit_team_motd(self, team_id: UUID, message: str):
        user: User = g.user
        try:
            team: Team = Team.get(id=team_id)
            if team.leader != user.id:
                raise TeamManagerException(f"Cannot edit team \"{team.name}\"'s MOTD as it doesn't belong to the user.")
            team.motd = message

            team.save()
        except Team.DoesNotExist as exc:
            raise TeamManagerException(f"Team {team_id} doesn't exist!", team_id) from exc

    def delete_team(self, team_id: UUID):
        user: User = g.user
        try:
            team: Team = Team.get(id=team_id)
            if team.leader != user.id:
                raise TeamManagerException(f"Cannot delete team \"{team.name}\" as it doesn't belong to the user.")
            team.delete()
        except Team.DoesNotExist as exc:
            raise TeamManagerException(f"Team {team_id} doesn't exist!", team_id) from exc

    def get_users_teams(self, user_id: UUID) -> list[Team]:
        teams_containing_user = list(Team.all())
        users_teams = self.get_teams_for_user(user_id)
        created_teams_ids = [team.id for team in users_teams]
        return [*users_teams, *[team for team in teams_containing_user if team.id not in created_teams_ids]]



import json
import logging
from glob import glob, escape
from pathlib import Path
from typing import TypedDict
from xml.etree import ElementTree

from argus.backend.plugins.sirenada.types import RawSirenadaRequest, RawSirenadaTestCase

from argus.client.base import ArgusClient


LOGGER = logging.getLogger(__name__)

class SirenadaEnv(TypedDict):
    SIRENADA_JOB_ID: str
    SIRENADA_BROWSER: str
    SIRENADA_REGION: str
    SIRENADA_CLUSTER: str
    WORKSPACE: str
    BUILD_NUMBER: str
    BUILD_URL: str
    JOB_NAME: str


class TestCredentials(TypedDict):
    SIRENADA_TEST_ID: str
    SIRENADA_USER_NAME: str
    SIRENADA_USER_PASS: str
    SIRENADA_OTP_SECRET: str
    ClusterID: str
    region: str

class ArgusSirenadaClient(ArgusClient):
    test_type = "sirenada"
    schema_version: None = "v1"
    _junit_xml_filename = "junit_results.xml"
    _credentials_filename = "test_credentials.json"
    _bucket_url_template = "https://sirenada-results.s3.us-east-1.amazonaws.com/{s3_id}/{test_results_dir}/{filename}"
    TEST_TAG_MAPPING = {
        "failure": "failed",
        "error": "error",
        "skipped": "skipped"
    }

    def __init__(self, auth_token: str, base_url: str, api_version="v1") -> None:
        self.results_path: Path | None = None
        super().__init__(auth_token, base_url, api_version)

    def _verify_required_files_exist(self, results_path: Path):
        assert (results_path / self._junit_xml_filename).exists(), "Missing jUnit XML results file!"
        assert (results_path / self._credentials_filename).exists(), "Missing test credentials file!"

    @staticmethod
    def _verify_env(env: SirenadaEnv):
        for key in SirenadaEnv.__annotations__.keys():
            assert env.get(key), f"Missing required key {key} in the environment"

    def _set_files_for_failed_test(self, test_case: RawSirenadaTestCase):
        test_results_dir = f"sirenada-{test_case['sirenada_test_id']}"
        results_dir: Path = self.results_path / test_results_dir

        just_class_name = test_case.get("class_name").split(".")[-1]
        requests_pattern = f"{just_class_name}-{test_case.get('test_name')}.requests.json"
        screenshot_pattern = f"{just_class_name}-{escape(test_case.get('test_name'))}-*.png"

        if not results_dir.exists():
            return

        if not (s3_bucket_id_path := (results_dir / "s3_id")).exists():
            return

        with open(s3_bucket_id_path, "rt") as s3_id_file:
            test_case["s3_folder_id"] = s3_id_file.read().strip()

        test_case["requests_file"] = self._bucket_url_template.format(
            s3_id=test_case.get("s3_folder_id"),
            test_results_dir=test_results_dir,
            filename=requests_pattern,
        )

        screenshots = glob(pathname=screenshot_pattern, root_dir=results_dir)
        if len(screenshots) > 0:
            test_case["screenshot_file"] = self._bucket_url_template.format(
                s3_id=test_case.get("s3_folder_id"),
                test_results_dir=test_results_dir,
                filename=screenshots[0],
            )

    def _set_case_status(self, raw_case: RawSirenadaTestCase, junit_case: ElementTree.Element):
        subelements = list(junit_case.iter())
        if len(subelements) == 2:
            status_node = subelements[1]
            raw_case["status"] = self.TEST_TAG_MAPPING.get(status_node.tag, "failed")
            raw_case["message"] = status_node.get("message", "#NO_MESSAGE")
            raw_case["stack_trace"] = status_node.text
            self._set_files_for_failed_test(raw_case)
        else:
            raw_case["status"] = "passed"

    def _parse_junit_results(self, junit_xml: Path, credentials: TestCredentials, env: SirenadaEnv) -> list[RawSirenadaTestCase]:
        etree = ElementTree.parse(source=junit_xml)
        cases = list(etree.iter("testcase"))
        assert len(cases) > 0, "No test cases found in test results XML."

        parsed_cases: list[RawSirenadaTestCase] = []
        for case in cases:
            raw_case: RawSirenadaTestCase = {}
            raw_case["class_name"] = case.get("classname", "#NO_CLASSNAME")
            raw_case["file_name"] = case.get("file", "#NO_FILE")
            raw_case["test_name"] = case.get("name", "#NO_TEST_NAME")
            raw_case["browser_type"] = env.get("SIRENADA_BROWSER")
            raw_case["cluster_type"] = env.get("SIRENADA_CLUSTER")
            raw_case["duration"] = float(case.attrib.get("time", "0.0"))
            raw_case["sirenada_test_id"] = credentials.get("SIRENADA_TEST_ID")
            raw_case["sirenada_user"] = credentials.get("SIRENADA_USER_NAME")
            raw_case["sirenada_password"] = credentials.get("SIRENADA_USER_PASS")
            self._set_case_status(raw_case, case)
            parsed_cases.append(raw_case)

        return parsed_cases

    @staticmethod
    def _read_credentials(credentials_path: Path) -> TestCredentials:
        with open(credentials_path, "rt") as creds_file:
            credentials: TestCredentials = json.load(creds_file)

        assert credentials.get("SIRENADA_TEST_ID"), "Credentials are missing required field: SIRENADA_TEST_ID"
        return credentials

    def submit_sirenada_run(self, env: SirenadaEnv, test_results_path: Path):
        self.results_path = test_results_path
        try:
            self._verify_env(env)
            self._verify_required_files_exist(self.results_path)
        except AssertionError as exc:
            LOGGER.critical("%s", exc.args[0])
            raise exc

        credentials = self._read_credentials(self.results_path / self._credentials_filename)
        results = self._parse_junit_results(junit_xml=self.results_path / self._junit_xml_filename, credentials=credentials, env=env)
        request_body: RawSirenadaRequest = {}

        request_body["run_id"] = env.get("SIRENADA_JOB_ID")
        request_body["build_id"] = env.get("JOB_NAME")
        request_body["build_job_url"] = env.get("BUILD_URL")
        request_body["region"] = env.get("SIRENADA_REGION")
        request_body["results"] = results

        response = self.submit_run(run_type=self.test_type, run_body=request_body)
        self.check_response(response)



from flask import Blueprint, request

from argus.backend.error_handlers import handle_api_exception
from argus.backend.service.user import api_login_required
from argus.backend.plugins.driver_matrix_tests.service import DriverMatrixService

bp = Blueprint("driver_matrix_api", __name__, url_prefix="/driver_matrix")
bp.register_error_handler(Exception, handle_api_exception)


@bp.route("/test_report", methods=["GET"])
@api_login_required
def driver_matrix_test_report():

    build_id = request.args.get("buildId")
    if not build_id:
        raise Exception("No build id provided")
    

    result = DriverMatrixService().tested_versions_report(build_id=build_id)
    return {
        "status": "ok",
        "response": result
    }



from pathlib import Path
import typing
from argus.backend.plugins.core import PluginInfoBase
from argus.backend.plugins.core import PluginModelBase


class PluginModule(typing.Protocol):
    # pylint: disable=too-few-public-methods
    PluginInfo: PluginInfoBase


def plugin_loader() -> dict[str, PluginInfoBase]:
    loader_path = Path(__file__).parent
    plugin_dirs = [p for p in loader_path.glob("*") if p.is_dir()]
    modules = {}
    parent_module = ".".join(__name__.split(".")[:-1])
    for directory in plugin_dirs:
        if (plugin_path := directory / "plugin.py").exists():
            rel_path = str(plugin_path.relative_to(loader_path))
            module_path = rel_path.replace("/", ".").replace(".py", "")
            module: PluginModule = __import__(
                f"{parent_module}.{module_path}",
                globals=globals(),
                fromlist=("PluginInfo",)
            )
            plugin = module.PluginInfo
            modules[plugin.name] = plugin

    return modules


AVAILABLE_PLUGINS = plugin_loader()


def all_plugin_models() -> list[PluginModelBase]:
    return [model for plugin in AVAILABLE_PLUGINS.values() for model in plugin.all_models]


def all_plugin_types():
    return [user_type for plugin in AVAILABLE_PLUGINS.values() for user_type in plugin.all_types]
